import tempfile
from copy import deepcopy
from test.conftest import test_config

from pptagent.presentation import Presentation
from pptagent.utils import Config


def test_presentation():
    presentation = Presentation.from_file(test_config.ppt, Config(tempfile.mkdtemp()))
    assert len(presentation.slides) == 1
    for sld in presentation.slides:
        sld.to_html(show_image=False)
    deepcopy(presentation)
    presentation.save("test.pptx", layout_only=True)



from typing import Literal

from pydantic import BaseModel, Field, create_model


class SlideElement(BaseModel):
    name: str
    data: list[str]

    @classmethod
    def response_model(cls, elements: list[str]):
        return create_model(
            cls.__name__,
            name=(Literal[*elements], Field(...)),  # type: ignore
            data=(list[str], Field(...)),
            __base__=BaseModel,
        )


class EditorOutput(BaseModel):
    elements: list[SlideElement]

    @property
    def dict(self):
        return {element.name: element.data for element in self.elements}

    @classmethod
    def response_model(cls, elements: list[str]):
        return create_model(
            cls.__name__,
            elements=(list[SlideElement.response_model(elements)], Field(...)),
            __base__=BaseModel,
        )

    def __contains__(self, key: str):
        for element in self.elements:
            if element.name == key:
                return True
        return False

    def __getitem__(self, key: str):
        for element in self.elements:
            if element.name == key:
                return element
        raise KeyError(f"Element '{key}' not found")


class LayoutChoice(BaseModel):
    reasoning: str
    layout: str

    @classmethod
    def response_model(cls, layouts: list[str]):
        return create_model(
            cls.__name__,
            reasoning=(str, Field(...)),
            layout=(Literal[*layouts], Field(...)),  # type: ignore
            __base__=BaseModel,
        )


class TemplateChoice(BaseModel):
    reasoning: str
    template: str

    @classmethod
    def response_model(cls, templates: list[str]):
        return create_model(
            cls.__name__,
            reasoning=(str, Field(...)),
            template=(Literal[*templates], Field(...)),  # type: ignore
            __base__=BaseModel,
        )



import asyncio
import os
import re
from contextlib import AsyncExitStack

from jinja2 import Environment, StrictUndefined
from pydantic import BaseModel, Field, create_model

from pptagent.agent import Agent
from pptagent.llms import AsyncLLM
from pptagent.model_utils import language_id
from os.path import basename, exists, join
from pptagent.utils import (
    Language,
    get_logger,
    package_join,
)

from .doc_utils import (
    get_tree_structure,
    process_markdown_content,
    split_markdown_by_headings,
)
from .element import Media, Metadata, Section, SubSection, Table, link_medias

logger = get_logger(__name__)

env = Environment(undefined=StrictUndefined)

MERGE_METADATA_PROMPT = env.from_string(
    open(
        package_join("prompts", "document", "merge_metadata.txt"), encoding="utf-8"
    ).read()
)
LITERAL_CONSTRAINT = os.getenv("LITERAL_CONSTRAINT", "false").lower() == "true"


class Document(BaseModel):
    image_dir: str
    language: Language
    metadata: dict[str, str]
    sections: list[Section]

    def validate_medias(self, image_dir: str | None = None):
        """Validate and fix media file paths"""
        if image_dir is not None:
            self.image_dir = image_dir
        assert exists(self.image_dir), f"image directory is not found: {self.image_dir}"
        for media in self.iter_medias():
            if exists(media.path):
                continue
            base_name = basename(media.path)
            if exists(join(self.image_dir, base_name)):
                media.path = join(self.image_dir, base_name)
            else:
                raise FileNotFoundError(f"image file not found: {media.path}")

    def get_overview(self, include_summary: bool = False, include_image: bool = True):
        """Get document overview with sections and subsections"""
        overview = ""
        for section in self.sections:
            overview += f"<section>{section.title}</section>\n"
            if include_summary:
                overview += f"\tSummary: {section.summary}\n"
            for subsection in section.content:
                if isinstance(subsection, SubSection):
                    overview += f"\t<subsection>{subsection.title}</subsection>\n"
                elif include_image and isinstance(subsection, Media):
                    overview += (
                        f"\t<image>{subsection.path}</image>: {subsection.caption}\n"
                    )
            overview += "\n"
        return overview

    def iter_medias(self):
        """Iterate over all media items in the document"""
        for section in self.sections:
            yield from section.iter_medias()

    def find_media(self, caption: str | None = None, path: str | None = None):
        """Find media by caption or path"""
        for media in self.iter_medias():
            if caption is not None and media.caption == caption:
                return media
            if path is not None and media.path == path:
                return media
        raise ValueError(f"Image caption or path not found: {caption} or {path}")

    @classmethod
    async def _parse_chunk(
        cls,
        extractor: Agent,
        markdown_chunk: str,
        image_dir: str,
        language_model: AsyncLLM,
        vision_model: AsyncLLM,
        limiter: asyncio.Semaphore | AsyncExitStack,
    ):
        markdown, medias = process_markdown_content(
            markdown_chunk,
        )
        async with limiter:
            _, section = await extractor(
                markdown_document=markdown,
                response_format=Section.response_model(),
            )
            metadata = section.pop("metadata", {})
            section["content"] = section.pop("subsections")
            section = Section(**section, markdown_content=markdown_chunk)
            link_medias(medias, section)
            async with asyncio.TaskGroup() as tg:
                for media in section.iter_medias():
                    media.parse(image_dir)
                    if isinstance(media, Table):
                        tg.create_task(media.get_caption(language_model))
                    else:
                        tg.create_task(media.get_caption(vision_model))
        return metadata, section

    @classmethod
    async def from_markdown(
        cls,
        markdown_content: str,
        language_model: AsyncLLM,
        vision_model: AsyncLLM,
        image_dir: str,
        max_at_once: int | None = None,
    ):
        doc_extractor = Agent(
            "doc_extractor",
            llm_mapping={"language": language_model, "vision": vision_model},
        )
        document_tree = get_tree_structure(markdown_content)
        headings = re.findall(r"^#+\s+.*", markdown_content, re.MULTILINE)
        splited_chunks = await split_markdown_by_headings(
            markdown_content, headings, document_tree, language_model
        )

        metadata = []
        sections = []
        tasks = []

        limiter = (
            asyncio.Semaphore(max_at_once)
            if max_at_once is not None
            else AsyncExitStack()
        )
        async with asyncio.TaskGroup() as tg:
            for chunk in splited_chunks:
                tasks.append(
                    tg.create_task(
                        cls._parse_chunk(
                            doc_extractor,
                            chunk,
                            image_dir,
                            language_model,
                            vision_model,
                            limiter,
                        )
                    )
                )

        # Process results in order
        for task in tasks:
            meta, section = task.result()
            metadata.append(meta)
            sections.append(section)

        merged_metadata = await language_model(
            MERGE_METADATA_PROMPT.render(metadata=metadata),
            return_json=True,
            response_format=create_model(
                "MetadataList",
                metadata=(list[Metadata], Field(...)),
                __base__=BaseModel,
            ),
        )
        metadata = {meta["name"]: meta["value"] for meta in merged_metadata["metadata"]}
        return cls(
            image_dir=image_dir,
            language=language_id(markdown_content),
            metadata=metadata,
            sections=sections,
        )

    def index(self, target_item: SubSection | Media | Table):
        """Get the index position of a content item"""
        for i, (_, content) in enumerate(self):
            if content is target_item:
                return i
        raise ValueError("Item not found in document")

    def pop(self, index: int):
        """Remove and return content item at specified index position"""
        for idx, (section, content) in enumerate(self):
            if idx == index:
                return section.content.pop(section.content.index(content))
        raise IndexError("Index out of range")

    def insert(self, item: SubSection | Media | Table, target_index: int):
        """Insert content item after the specified index position"""
        for idx, (section, content) in enumerate(self):
            if idx == target_index:
                section.content.insert(section.content.index(content), item)
                return

        self.sections[-1].content.append(item)

    def remove(self, target_item: SubSection | Media | Table):
        """Remove content item from document"""
        for section, content in self:
            if content is target_item:
                section.content.remove(target_item)
                return
        raise ValueError("Item not found in document")

    def __contains__(self, key: str):
        for section in self.sections:
            if section.title == key:
                return True
        return False

    def __iter__(self):
        for section in self.sections:
            for content in section.content:
                yield section, content

    def __getitem__(self, key: int | slice | str):
        """Get content item by index, slice or section title"""
        if isinstance(key, slice):
            return [content for _, content in list(self)[key]]
        else:
            for i, (sec, content) in enumerate(self):
                if i == key:
                    return content
                elif sec.title == key:
                    return sec
            raise IndexError(f"Index out of range: {key}")

    @property
    def metainfo(self):
        return "\n".join([f"{k}: {v}" for k, v in self.metadata.items()])



import asyncio
from math import ceil
from typing import Literal

from jinja2 import StrictUndefined, Template
from pydantic import BaseModel, field_validator

from pptagent.llms import AsyncLLM
from pptagent.response import EditorOutput
from os.path import exists
from pptagent.utils import edit_distance, get_logger, package_join

logger = get_logger(__name__)

LENGTHY_REWRITE_PROMPT = Template(
    open(package_join("prompts", "lengthy_rewrite.txt"), encoding="utf-8").read(),
    undefined=StrictUndefined,
)


class Element(BaseModel):
    name: str
    data: list[str]
    type: Literal["text", "image"]
    suggested_characters: int | None = None
    variable_length: tuple[int, int] | None = None
    variable_data: dict[str, list[str]] | None = None

    def model_post_init(self, _):
        if self.type == "text":
            self.suggested_characters = max(len(i) for i in self.data)

    def get_schema(self):
        schema = f"Element: {self.name}\n"
        schema += f"\type: {self.type}\n"
        if self.type == "text":
            schema += f"\tsuggested_characters: {self.suggested_characters}\n"
        if self.variable_length is not None:
            schema += f"\tThe length of the element can vary between {self.variable_length[0]} and {self.variable_length[1]}\n"
        schema += f"\tThe default quantity of the element is {len(self.data)}\n"
        return schema


class Layout(BaseModel):
    title: str
    template_id: int
    slides: list[int]
    elements: list[Element]
    vary_mapping: dict[int, int] | None = None  # mapping for variable elements

    @field_validator("elements")
    def validate_elements(cls, v):
        num_vary_elements = sum((el.variable_length is not None) for el in v)
        if num_vary_elements > 1:
            raise ValueError("Only one variable element allowed in a layout")
        return v

    def index_template_slide(self, data: EditorOutput):
        old_data = {}
        template_id = self.template_id

        for el in self.elements:
            if el.variable_length is not None:
                num_vary = len(data[el.name].data)
                if not (el.variable_length[0] <= num_vary <= el.variable_length[1]):
                    raise ValueError(
                        f"The length of {el.name}: {num_vary} is not within the allowed range: [{el.variable_length[0]}, {el.variable_length[1]}]"
                    )
                template_id = self.vary_mapping[str(num_vary)]
                key = str(len(data[el.name].data))
                old_data[el.name] = el.variable_data[key]
            else:
                old_data[el.name] = el.data

        return template_id, old_data

    def validate(self, editor_output: EditorOutput, allowed_images: list[str]):
        for el in self.elements:
            if el.name not in editor_output:
                raise ValueError(f"Element {el.name} not found in editor output")
            if self[el.name].type != "image":
                continue
            assert len(editor_output[el.name].data) == 0 or len(allowed_images) > 0, (
                "No images provided for slide generation, please leave a blank list for this element"
            )
            for i in range(len(editor_output[el.name].data)):
                sim_image = max(
                    allowed_images,
                    key=lambda x: edit_distance(x, editor_output[el.name].data[i]),
                )
                if edit_distance(
                    sim_image, editor_output[el.name].data[i]
                ) < 0.5 or not exists(sim_image):
                    raise ValueError(
                        f"Image {editor_output[el.name].data[i]} not found\n"
                        "Please check the image path and use only existing images\n"
                        "Or, leave a blank list for this element"
                    )
                editor_output[el.name].data[i] = sim_image

    async def length_rewrite(
        self,
        editor_output: EditorOutput,
        length_factor: float,
        language_model: AsyncLLM,
    ):
        async with asyncio.TaskGroup() as tg:
            tasks = []
            for el in editor_output.elements:
                if self[el.name].type != "text":
                    continue
                charater_counts = max([len(i) for i in el.data])
                expected_length = ceil(
                    self[el.name].suggested_characters * length_factor
                )
                if charater_counts - expected_length > 5:
                    task = tg.create_task(
                        language_model(
                            LENGTHY_REWRITE_PROMPT.render(
                                el_name=el.name,
                                content=el.data,
                                suggested_characters=f"{self[el.name].suggested_characters} characters",
                            ),
                            return_json=True,
                        )
                    )
                    tasks.append([el.name, task])

            for el_name, task in tasks:
                editor_output[el_name].data = await task

    @property
    def content_schema(self):
        return "\n".join([el.get_schema() for el in self.elements])

    def remove_item(self, item: str):
        for el in self.elements:
            if item in el.data:
                el.data.remove(item)
                if len(el.data) == 0:
                    self.elements.remove(el)
                return
        else:
            raise ValueError(f"Item {item} not found in layout {self.title}")

    def __contains__(self, key: str | int):
        if isinstance(key, int):
            return key in self.slides
        elif isinstance(key, str):
            for el in self.elements:
                if el.name == key:
                    return True
            return False
        raise ValueError(f"Invalid key type: {type(key)}, should be str or int")

    def __getitem__(self, key: str):
        for el in self.elements:
            if el.name == key:
                return el
        raise ValueError(f"Element {key} not found")

    def __iter__(self):
        return iter(self.elements)

    def __len__(self):
        return len(self.elements)



import os
import tempfile
from test.conftest import test_config

import pytest

from pptagent.utils import get_json_from_response, package_join, ppt_to_images


def test_package_data():
    """Test package data."""
    assert len(os.listdir(package_join("resource"))) == 1
    assert len(os.listdir(package_join("prompts"))) > 0
    assert len(os.listdir(package_join("roles"))) > 0


def test_extract_json_from_markdown_block():
    """Test extracting JSON from a markdown code block."""
    response = """
    Here's the JSON you requested:

    ```json
    {
        "name": "John",
        "age": 30,
        "city": "New York"
    }
    ```

    Let me know if you need anything else.
    """

    result = get_json_from_response(response)
    assert isinstance(result, dict)
    assert result["name"] == "John"
    assert result["age"] == 30
    assert result["city"] == "New York"


def test_extract_list_json():
    """Test extracting JSON with a list of objects."""
    response = """
    Here's the JSON you requested:
    ```json
    [
        {"name": "John", "age": 30},
        {"name": "Jane", "age": 25}
    ]
    ```
    """

    result = get_json_from_response(response)
    assert result == [{"name": "John", "age": 30}, {"name": "Jane", "age": 25}]


def test_extract_complex_json():
    """Test extracting JSON with minor syntax errors that can be repaired."""
    response = """
```json\n{\n    "table_data": [\n        ["Domain", "", "", "Presentation", "", ""],\n        ["", "#Chars", "#Figs", "#Chars", "#Figs", "#Pages"],\n        ["Culture", "12,708", "2.9", "6,585", "12.8", "14.3"],\n        ["Education", "12,305", "5.5", "3,993", "12.9", "13.9"],\n        ["Science", "16,661", "4.8", "5,334", "24.0", "18.4"],\n        ["Society", "13,019", "7.3", "3,723", "9.8", "12.9"],\n        ["Tech", "18,315", "11.4", "5,325", "12.9", "16.8"]\n    ],\n    "merge_area": [\n        [0, 0, 0, 5],\n        [1, 0, 1, 1],\n        [1, 3, 1, 5]\n    ]\n}\n```
    """

    result = get_json_from_response(response)
    assert isinstance(result, dict)
    assert result["table_data"] == [
        ["Domain", "", "", "Presentation", "", ""],
        ["", "#Chars", "#Figs", "#Chars", "#Figs", "#Pages"],
        ["Culture", "12,708", "2.9", "6,585", "12.8", "14.3"],
        ["Education", "12,305", "5.5", "3,993", "12.9", "13.9"],
        ["Science", "16,661", "4.8", "5,334", "24.0", "18.4"],
        ["Society", "13,019", "7.3", "3,723", "9.8", "12.9"],
        ["Tech", "18,315", "11.4", "5,325", "12.9", "16.8"],
    ]
    assert result["merge_area"] == [[0, 0, 0, 5], [1, 0, 1, 1], [1, 3, 1, 5]]


def test_extract_nested_json():
    """Test extracting nested JSON objects."""
    response = """
    Here's the JSON:

    {
        "person": {
            "name": "John",
            "age": 30
        },
        "address": {
            "city": "New York",
            "zip": "10001"
        }
    }
    """

    result = get_json_from_response(response)
    assert isinstance(result, dict)
    assert result["person"]["name"] == "John"
    assert result["address"]["city"] == "New York"


def test_json_not_found():
    """Test that an exception is raised when JSON is not found."""
    response = "This is just plain text with no JSON."

    with pytest.raises(Exception) as excinfo:
        get_json_from_response(response)

    assert "JSON not found" in str(excinfo.value)


def test_ppt_to_images_conversion():
    """Test converting a PPTX file to images."""
    # Run the conversion
    ppt_to_images(test_config.ppt, tempfile.mkdtemp())



import json
import os
from collections import defaultdict
from glob import glob

from jinja2 import Template
from tqdm.asyncio import tqdm

from .model_utils import ModelManager
from .presentation import Presentation
from os.path import dirname, exists, join
from .utils import Config, package_join

manager = ModelManager()
language_model = manager.language_model
vision_model = manager.vision_model

text_scorer = Template(
    open(
        package_join("prompts", "ppteval", "ppteval_content.txt"), encoding="utf-8"
    ).read()
)
vision_scorer = Template(
    open(
        package_join("prompts", "ppteval", "ppteval_style.txt"), encoding="utf-8"
    ).read()
)
style_descriptor = open(
    package_join("prompts", "ppteval", "ppteval_describe_style.txt"), encoding="utf-8"
).read()
content_descriptor = open(
    package_join("prompts", "ppteval", "ppteval_describe_content.txt"), encoding="utf-8"
).read()
ppt_extractor = Template(
    open(
        package_join("prompts", "ppteval", "ppteval_extract.txt"), encoding="utf-8"
    ).read()
)
logic_scorer = Template(
    open(
        package_join("prompts", "ppteval", "ppteval_coherence.txt"), encoding="utf-8"
    ).read()
)


def get_eval(prs_source: str):
    evals = defaultdict(dict)
    eval_file = join(dirname(prs_source), "evals.json")
    if exists(eval_file):
        with open(eval_file, encoding="utf-8") as f:
            evals |= json.load(f)
    return evals, eval_file


async def slide_score(prs_source: str, slide_folder: str):
    evals, eval_file = get_eval(prs_source)
    for slide_image in glob(join(slide_folder, "slide_*.jpg")) + glob(
        join(slide_folder, "slide_images", "slide_*.jpg")
    ):
        slide_descr = slide_image.replace(".jpg", ".json")
        if not os.path.exists(slide_descr):
            style_descr = await vision_model(style_descriptor, slide_image)
            content_descr = await vision_model(content_descriptor, slide_image)
            with open(slide_descr, "w", encoding="utf-8") as f:
                json.dump(
                    {"content": content_descr, "style": style_descr},
                    f,
                    indent=2,
                )
        else:
            descr = json.load(open(slide_descr, encoding="utf-8"))
            style_descr = descr["style"]
            content_descr = descr["content"]
        if slide_image not in evals["vision"]:
            evals["vision"][slide_image] = await language_model(
                vision_scorer.render(descr=style_descr), return_json=True
            )
        if slide_image not in evals["content"]:
            evals["content"][slide_image] = await language_model(
                text_scorer.render(descr=content_descr), return_json=True
            )
    with open(eval_file, "w", encoding="utf-8") as f:
        json.dump(evals, f, indent=2)


async def pres_score(prs_source: str):
    tmp_config = Config(dirname(prs_source))
    evals, eval_file = get_eval(prs_source)
    if "logic" in evals:
        return
    slide_descr = join(prs_source.replace(".pptx", ""), "extracted.json")
    if not exists(slide_descr):
        presentation = Presentation.from_file(prs_source, tmp_config).to_text()
        extracted = language_model(
            ppt_extractor.render(presentation=presentation),
            return_json=True,
        )
        with open(slide_descr, "w", encoding="utf-8") as f:
            json.dump(extracted, f, indent=2)
    else:
        extracted = json.load(open(slide_descr, encoding="utf-8"))
    evals["logic"] = await language_model(
        logic_scorer.render(
            presentation=extracted,
        ),
        return_json=True,
    )
    with open(eval_file, "w", encoding="utf-8") as f:
        json.dump(evals, f, indent=2)


async def eval_ppt(prs_files: list[str], slide_folders: list[str]):
    await tqdm.gather(*[pres_score(prs_file) for prs_file in prs_files])
    await tqdm.gather(
        *[
            slide_score(prs_file, slide_folder)
            for prs_file, slide_folder in zip(prs_files, slide_folders)
        ]
    )

    # Calculate average scores directly in this function
    all_scores = {"vision": [], "content": [], "logic": []}

    for prs_file in prs_files:
        evals, _ = get_eval(prs_file)

        # Vision scores
        vision_scores = [score.get("score", 0) for score in evals["vision"].values()]
        if vision_scores:
            all_scores["vision"].extend(vision_scores)

        # Content scores
        content_scores = [score.get("score", 0) for score in evals["content"].values()]
        if content_scores:
            all_scores["content"].extend(content_scores)

        # Logic score
        if "logic" in evals and "score" in evals["logic"]:
            all_scores["logic"].append(evals["logic"]["score"])

    # Calculate averages
    avg_scores = {}
    for category, scores in all_scores.items():
        if scores:
            avg_scores[category] = sum(scores) / len(scores)
        else:
            avg_scores[category] = 0

    # Display results
    print("\n=== Average Scores by Dimension ===")
    for category, score in avg_scores.items():
        print(f"{category.capitalize()}: {score:.2f}")

    # Calculate overall average
    if avg_scores:
        overall_avg = sum(avg_scores.values()) / len(avg_scores)
        print(f"Overall: {overall_avg:.2f}")

    return avg_scores



import os
import re
from contextvars import ContextVar

from bs4 import BeautifulSoup
from jinja2 import Environment, StrictUndefined
from pydantic import BaseModel

from pptagent.llms import AsyncLLM
from pptagent.utils import edit_distance, package_join

env = Environment(undefined=StrictUndefined)


MARKDOWN_IMAGE_REGEX = re.compile(r"!\[.*\]\(.*\)")
MARKDOWN_TABLE_REGEX = re.compile(
    r"(\|.*\|)|((<html><body>)?<table>.*</table>(</body></html>)?)"
)
HEADING_EXTRACT_PROMPT = env.from_string(
    open(
        package_join("prompts", "document", "heading_extract.txt"), encoding="utf-8"
    ).read()
)

MIN_CHUNK_SIZE: int = os.getenv("MIN_CHUNK_SIZE", 512)
MAX_CHUNK_SIZE: int = os.getenv("MAX_CHUNK_SIZE", 32768)


def count_markdown_chunks(markdown_text):
    """
    Count characters in each heading chunk of a Markdown document

    Args:
        markdown_text (str): Markdown text content

    Returns:
        list: List containing heading information and character counts
    """
    lines = markdown_text.split("\n")
    chunks = []
    current_heading = None
    current_content = []
    current_level = 0

    for line in lines:
        # Check if the line is a heading
        heading_match = re.match(r"^(#{1,6})\s+(.+)", line)

        if heading_match:
            # Save the previous chunk if exists
            if current_heading is not None:
                content_text = "\n".join(current_content)
                char_count = len(content_text.strip())
                chunks.append(
                    {
                        "level": current_level,
                        "heading": current_heading,
                        "char_count": char_count,
                        "content": content_text.strip(),
                    }
                )

            # Start a new chunk
            current_level = len(heading_match.group(1))
            current_heading = heading_match.group(2).strip()
            current_content = []
        else:
            # Non-heading line, add to current content
            current_content.append(line)

    # Handle the last chunk
    if current_heading is not None:
        content_text = "\n".join(current_content)
        char_count = len(content_text.strip())
        chunks.append(
            {
                "level": current_level,
                "heading": current_heading,
                "char_count": char_count,
                "content": content_text.strip(),
            }
        )

    return chunks


def calculate_hierarchical_counts(chunks):
    """
    Calculate character statistics including parent-child relationships

    Args:
        chunks (list): Original chunk list

    Returns:
        list: Chunk list with hierarchical statistics
    """

    def get_children_count(parent_index, parent_level):
        """Recursively calculate total character count of child headings"""
        total = 0
        for i in range(parent_index + 1, len(chunks)):
            current_level = chunks[i]["level"]

            # Stop if encountering same or higher level heading
            if current_level <= parent_level:
                break

            # Count all child direct character counts
            if current_level > parent_level:
                total += chunks[i]["char_count"]

        return total

    # Add hierarchical statistics to each chunk
    for i, chunk in enumerate(chunks):
        # Direct content character count
        chunk["direct_char_count"] = chunk["char_count"]

        # Children content character count
        children_count = get_children_count(i, chunk["level"])
        chunk["children_char_count"] = children_count

        # Total character count (self + all children)
        chunk["total_char_count"] = chunk["char_count"] + children_count

    return chunks


def display_results(chunks):
    """
    Format and display statistics results

    Args:
        chunks (list): Heading chunk list
    """
    print("Markdown Heading Character Statistics:")
    print("=" * 80)

    total_chars = 0
    root_total = 0

    for i, chunk in enumerate(chunks, 1):
        indent = "  " * (chunk["level"] - 1)
        heading_prefix = "#" * chunk["level"]

        print(f"{i}. {indent}{heading_prefix} {chunk['heading']}")
        print(f"   {indent}├─ Direct content: {chunk['direct_char_count']} characters")

        if chunk["children_char_count"] > 0:
            print(
                f"   {indent}├─ Children content: {chunk['children_char_count']} characters"
            )
            print(f"   {indent}└─ Total: {chunk['total_char_count']} characters")
        else:
            print(f"   {indent}└─ Total: {chunk['total_char_count']} characters")

        total_chars += chunk["direct_char_count"]

        if chunk["level"] == 1:
            root_total += chunk["total_char_count"]

        print()

    print("=" * 80)
    print(f"Total document characters: {total_chars}")
    print(f"Root level total: {root_total}")


def get_tree_structure(markdown: str, add_tag: bool = True):
    """
    Display tree structure statistics

    Args:
        markdown (str): Markdown content
    """
    chunks = count_markdown_chunks(markdown.strip())
    chunks_with_hierarchy = calculate_hierarchical_counts(chunks)

    tree = ""
    for chunk in chunks_with_hierarchy:
        indent = "  " * (chunk["level"] - 1)
        tree_symbol = "├─" if chunk["level"] > 1 else "■"
        if add_tag:
            heading = f"<title>{chunk['heading']}</title>"
        else:
            heading = chunk["heading"]

        tree += (
            f"{indent}{tree_symbol} {heading} "
            f"[Direct:{chunk['direct_char_count']} | Total Characters:{chunk['total_char_count']}]\n"
        )

    return tree


def find_middle_heading_position(text: str) -> int:
    """
    Find a heading position near the middle of the text.

    Args:
        text (str): Markdown text to analyze

    Returns:
        int: Position index of the chosen heading, or -1 if no headings found
    """
    lines = text.splitlines()
    middle_char_pos = len(text) // 2

    # Find all headings and their positions
    heading_positions = []
    char_pos = 0

    for i, line in enumerate(lines):
        if re.match(r"^#{1,6}\s+.+", line):
            heading_positions.append((i, char_pos))
        char_pos += len(line)

    if len(heading_positions) < 2:
        return -1

    # Find heading closest to the middle
    closest_heading = min(heading_positions, key=lambda x: abs(x[1] - middle_char_pos))[
        0
    ]

    return closest_heading


def split_large_chunks(sections: list[str]) -> list[str]:
    """
    Split sections that exceed the maximum chunk size by finding a heading near the middle.

    Args:
        sections (list[str]): List of markdown sections

    Returns:
        list[str]: List of sections with large chunks split
    """
    result = []
    for section in sections:
        if len(section) <= MAX_CHUNK_SIZE:
            result.append(section)
            continue

        # Try to split the section at a heading near the middle
        lines = section.splitlines()
        middle_heading_idx = find_middle_heading_position(section)

        # If no headings found or can't split further, just keep as is
        if middle_heading_idx == -1:
            result.append(section)
            continue

        # Split the section at the middle heading
        first_part = "\n".join(lines[:middle_heading_idx]).strip()
        second_part = "\n".join(lines[middle_heading_idx:]).strip()

        # Recursively split if still too large
        result.extend(split_large_chunks([first_part, second_part]))

    return result


# global context variable for allowed headings, used to validate headings in async context
_allowed_headings: ContextVar[list[str]] = ContextVar("allowed_headings", default=[])


class LogicHeadings(BaseModel):
    headings: list[str]

    def model_post_init(self, _):
        self.headings = [
            max(_allowed_headings.get(), key=lambda x: edit_distance(x, h))
            for h in self.headings
        ]

    @classmethod
    def response_model(cls, allowed_headings: list[str]):
        _allowed_headings.set(allowed_headings)
        return cls


async def split_markdown_by_headings(
    markdown_content: str,
    headings: list[str],
    document_tree: str,
    language_model: AsyncLLM,
) -> list[str]:
    """
    Split markdown content using headings as separators.

    Args:
        markdown_content (str): The markdown content to split
        headings (list[str]): List of heading strings to split by
        adjusted_headings (list[str], optional): List of adjusted heading strings
        min_chunk_size (int, optional): Minimum chunk size. Defaults to 64.

    Returns:
        list[str]: List of content sections
    """
    if len(headings) < 4:
        logic_headings = headings
    else:
        logic_headings = await language_model(
            HEADING_EXTRACT_PROMPT.render(tree=document_tree),
            return_json=True,
            response_format=LogicHeadings.response_model(headings),
        )
        logic_headings = LogicHeadings(**logic_headings).headings

    sections = []
    current_section = []

    for line in markdown_content.splitlines():
        if any(line.startswith(h) for h in logic_headings):
            if len(current_section) != 0:
                sections.append("\n".join(current_section).strip())
            current_section = [line]
        else:
            current_section.append(line)

    if len(current_section) != 0:
        sections.append("\n".join(current_section).strip())

    # if a chunk is too small, merge it with the previous chunk
    for i in reversed(range(1, len(sections))):
        if len(sections[i]) < MIN_CHUNK_SIZE:
            sections[i - 1] += "\n\n" + sections.pop(i)

    if len(sections) > 1 and len(sections[0]) < MIN_CHUNK_SIZE:
        sections[0] += "\n\n" + sections.pop(1)

    # Split sections that exceed MAX_CHUNK_SIZE
    sections = split_large_chunks(sections)

    return sections


def process_markdown_content(
    markdown_content: str,
    max_chunk_size: int = 256,
):
    """
    Process markdown content into paragraphs and media elements.

    Args:
        markdown_content (str): The original markdown text
        max_chunk_size (int, optional): Maximum chunk size. Defaults to 256.
        table_regex (Pattern, optional): Regex to identify tables
        image_regex (Pattern, optional): Regex to identify images

    Returns:
        list: List of media elements with their context
    """
    paragraphs = []
    medias_chunks = []

    # Split text into paragraphs and identify media elements
    for i, para in enumerate(markdown_content.split("\n\n")):
        para = para.strip()
        if not para:
            continue

        paragraph = {"markdown_content": para, "index": i}

        if MARKDOWN_TABLE_REGEX.match(para):
            paragraph["type"] = "table"
            medias_chunks.append(paragraph)
        elif MARKDOWN_IMAGE_REGEX.match(para):
            paragraph["type"] = "image"
            medias_chunks.append(paragraph)
        else:
            paragraphs.append(paragraph)

    # Add context to each media element
    for media in medias_chunks:
        pre_chunk = ""
        after_chunk = ""

        # Get preceding context
        for chunk in paragraphs[: media["index"]]:
            pre_chunk += chunk["markdown_content"] + "\n\n"
            if len(pre_chunk) > max_chunk_size:
                break

        # Get following context
        for chunk in paragraphs[media["index"] + 1 :]:
            after_chunk += chunk["markdown_content"] + "\n\n"
            if len(after_chunk) > max_chunk_size:
                break

        media["near_chunks"] = (pre_chunk, after_chunk)

    cleaned_markdown = "\n\n".join([p["markdown_content"] for p in paragraphs])
    return cleaned_markdown, medias_chunks


def parse_table_with_merges(
    html: str,
) -> tuple[list[list[str]], list[tuple[int, int, int, int]]]:
    """parse table in html with merge cell

    Args:
        html (str)

    Returns:
        cell_and_merge (cell: list[list[str]], merges: list[(x0: int, y0: int, x1: int, y1: int)])
    """
    soup = BeautifulSoup(html, "html.parser")
    table = soup.find("table")
    # Calculate maximum rows and columns of the table
    rows = table.find_all("tr")
    max_row = 0
    col_counter = []
    for row_idx, row in enumerate(rows):
        col_span_sum = 0
        for cell in row.find_all(["td", "th"]):
            row_span = int(cell.get("rowspan", 1))
            col_span = int(cell.get("colspan", 1))
            max_row = max(max_row, row_idx + row_span)
            col_span_sum += col_span
        col_counter.append(col_span_sum)
    max_col = max(col_counter) if col_counter else 0

    # Initialize data containers
    grid = [["" for _ in range(max_col)] for _ in range(max_row)]
    occupied = [[False for _ in range(max_col)] for _ in range(max_row)]
    merges = []

    # Main parsing logic
    for row_idx, row in enumerate(rows):
        col_idx = 0
        for cell in row.find_all(["td", "th"]):
            # Skip occupied columns
            while col_idx < max_col and occupied[row_idx][col_idx]:
                col_idx += 1
            if col_idx >= max_col:
                break

            # Parse cell attributes
            row_span = int(cell.get("rowspan", 1))
            col_span = int(cell.get("colspan", 1))
            cell_value = cell.get_text(strip=True)

            # Record merge range (closed interval)
            x0, y0 = row_idx, col_idx
            x1 = min(row_idx + row_span - 1, max_row - 1)
            y1 = min(col_idx + col_span - 1, max_col - 1)
            if not (x0 == x1 and y0 == y1):
                merges.append((x0, y0, x1, y1))

            # Fill top-left cell
            grid[x0][y0] = cell_value

            # Mark merged area as occupied
            for r in range(x0, x1 + 1):
                for c in range(y0, y1 + 1):
                    if r < max_row and c < max_col:
                        occupied[r][c] = True

            col_idx += col_span  # Move to next column
    return grid, merges



import os
from copy import deepcopy
from glob import glob
import tempfile
import zipfile

import aiofiles
import aiohttp
import numpy as np
import torch
import torchvision.transforms as T
from PIL import Image

from pptagent.llms import AsyncLLM
from os.path import join
from pptagent.utils import (
    Language,
    get_logger,
    is_image_path,
)

logger = get_logger(__name__)

# Lazy loading cache for the language ID model
_LID_MODEL = None


def _get_lid_model():
    from huggingface_hub.constants import HUGGINGFACE_HUB_CACHE

    """Get the language ID model, loading it lazily on first access."""
    global _LID_MODEL
    if _LID_MODEL is None:
        from fasttext import load_model
        from huggingface_hub import hf_hub_download

        lid_pattern = join(
            HUGGINGFACE_HUB_CACHE,
            "models--julien-c--fasttext-language-id",
            "*/*/lid.176.bin",
        )
        lid_files = glob(lid_pattern)
        if lid_files:
            _LID_MODEL = load_model(lid_files[0])
        else:
            _LID_MODEL = load_model(
                hf_hub_download(
                    repo_id="julien-c/fasttext-language-id",
                    filename="lid.176.bin",
                )
            )
    return _LID_MODEL


MINERU_API = os.environ.get("MINERU_API", None)
if MINERU_API is None:
    logger.warning("MINERU_API is not set, PDF parsing is not available")


class ModelManager:
    """
    A class to manage models.
    """

    def __init__(
        self,
        api_base: str | None = None,
        language_model_name: str | None = None,
        vision_model_name: str | None = None,
    ):
        """Initialize models from environment variables after instance creation"""
        if api_base is None:
            api_base = os.environ.get("API_BASE", None)
        if language_model_name is None:
            language_model_name = os.environ.get("LANGUAGE_MODEL", "gpt-4.1")
        if vision_model_name is None:
            vision_model_name = os.environ.get("VISION_MODEL", "gpt-4.1")
        self._image_model = None
        self._marker_model = None
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        self.language_model = AsyncLLM(language_model_name, api_base)
        self.vision_model = AsyncLLM(vision_model_name, api_base)

    @property
    def image_model(self):
        if self._image_model is None:
            self._image_model = get_image_model(device=self.device)
        return self._image_model

    async def test_connections(self) -> bool:
        """Test connections for all LLM models

        Returns:
            bool: True if all connections are successful, False otherwise
        """
        try:
            assert await self.language_model.test_connection()
            assert await self.vision_model.test_connection()
        except Exception as _:
            return False
        return True


def language_id(text: str) -> Language:
    model = _get_lid_model()
    return Language(
        lid=model.predict(text[:1024].replace("\n", ""))[0][0].replace("__label__", "")
    )


def get_image_model(device: str = None):
    from transformers import AutoModel, AutoProcessor

    """
    Initialize and return an image model and its feature extractor.

    Args:
        device (str): The device to run the model on.

    Returns:
        tuple: A tuple containing the feature extractor and the image model.
    """
    model_base = "google/vit-base-patch16-224-in21k"
    return (
        AutoProcessor.from_pretrained(
            model_base,
            torch_dtype=torch.float16,
            device_map=device,
            use_fast=True,
        ),
        AutoModel.from_pretrained(
            model_base,
            torch_dtype=torch.float16,
            device_map=device,
        ).eval(),
    )


async def parse_pdf(pdf_path: str, output_folder: str):
    """
    Parse a PDF file and extract text and images.

    Args:
        pdf_path (str): The path to the PDF file.
        output_path (str): The root directory to save the extracted content.

    Returns:
        str: The path to the extracted folder
    """
    assert MINERU_API is not None, "MINERU_API is not set"
    os.makedirs(output_folder, exist_ok=True)

    async with aiofiles.open(pdf_path, "rb") as f:
        pdf_content = await f.read()

    data = aiohttp.FormData()
    data.add_field(
        "files",
        pdf_content,
        filename=os.path.basename(pdf_path),
        content_type="application/pdf",
    )
    data.add_field("return_images", "True")
    data.add_field("response_format_zip", "True")

    async with aiohttp.ClientSession() as session:
        async with session.post(MINERU_API, data=data) as response:
            response.raise_for_status()
            content = await response.read()

            with tempfile.NamedTemporaryFile(delete=False, suffix=".zip") as tmp:
                tmp.write(content)
                zip_path = tmp.name

            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                top_level = {
                    name.split("/", 1)[0] for name in zip_ref.namelist() if name.strip()
                }
                if len(top_level) != 1:
                    raise RuntimeError("Expected exactly one top-level folder in zip")
                prefix = list(top_level)[0] + "/"

                for member in zip_ref.infolist():
                    filename = member.filename
                    dest_path = os.path.join(
                        output_folder, filename.removeprefix(prefix)
                    )

                    if not member.is_dir():
                        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
                        with zip_ref.open(member) as src, open(dest_path, "wb") as dst:
                            dst.write(src.read())


def get_image_embedding(
    image_dir: str, extractor, model, batchsize: int = 16
) -> dict[str, torch.Tensor]:
    """
    Generate image embeddings for images in a directory.

    Args:
        image_dir (str): The directory containing images.
        extractor: The feature extractor for images.
        model: The model used for generating embeddings.
        batchsize (int): The batch size for processing images.

    Returns:
        dict: A dictionary mapping image filenames to their embeddings.
    """
    transform = T.Compose(
        [
            T.Resize(int((256 / 224) * extractor.size["height"])),
            T.CenterCrop(extractor.size["height"]),
            T.ToTensor(),
            T.Normalize(mean=extractor.image_mean, std=extractor.image_std),
        ]
    )

    inputs = []
    embeddings = []
    images = [i for i in sorted(os.listdir(image_dir)) if is_image_path(i)]
    for file in images:
        image = Image.open(join(image_dir, file)).convert("RGB")
        inputs.append(transform(image))
        if len(inputs) % batchsize == 0 or file == images[-1]:
            batch = {"pixel_values": torch.stack(inputs).to(model.device)}
            embeddings.extend(model(**batch).last_hidden_state.detach())
            inputs.clear()
    return {image: embedding.flatten() for image, embedding in zip(images, embeddings)}


def images_cosine_similarity(embeddings: list[torch.Tensor]) -> torch.Tensor:
    """
    Calculate the cosine similarity matrix for a list of embeddings.
    Args:
        embeddings (list[torch.Tensor]): A list of image embeddings.

    Returns:
        torch.Tensor: A NxN similarity matrix.
    """
    embeddings = [embedding for embedding in embeddings]
    sim_matrix = torch.zeros((len(embeddings), len(embeddings)))
    for i in range(len(embeddings)):
        for j in range(i + 1, len(embeddings)):
            sim_matrix[i, j] = sim_matrix[j, i] = torch.cosine_similarity(
                embeddings[i], embeddings[j], -1
            )
    return sim_matrix


IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)


def average_distance(
    similarity: torch.Tensor, idx: int, cluster_idx: list[int]
) -> float:
    """
    Calculate the average distance between a point (idx) and a cluster (cluster_idx).

    Args:
        similarity (np.ndarray): The similarity matrix.
        idx (int): The index of the point.
        cluster_idx (list): The indices of the cluster.

    Returns:
        float: The average distance.
    """
    if idx in cluster_idx:
        return 0
    total_similarity = 0
    for idx_in_cluster in cluster_idx:
        total_similarity += similarity[idx, idx_in_cluster]
    return total_similarity / len(cluster_idx)


def get_cluster(similarity: np.ndarray, sim_bound: float = 0.65):
    """
    Cluster points based on similarity.

    Args:
        similarity (np.ndarray): The similarity matrix.
        sim_bound (float): The similarity threshold for clustering.

    Returns:
        list: A list of clusters.
    """
    sim_copy = deepcopy(similarity)
    num_points = sim_copy.shape[0]
    clusters = []
    added = [False] * num_points

    while True:
        max_avg_dist = sim_bound
        best_cluster = None
        best_point = None

        for c in clusters:
            for point_idx in range(num_points):
                if added[point_idx]:
                    continue
                avg_dist = average_distance(sim_copy, point_idx, c)
                if avg_dist > max_avg_dist:
                    max_avg_dist = avg_dist
                    best_cluster = c
                    best_point = point_idx

        if best_point is not None:
            best_cluster.append(best_point)
            added[best_point] = True
            sim_copy[best_point, :] = 0
            sim_copy[:, best_point] = 0
        else:
            if sim_copy.max() < sim_bound:
                # append the remaining points individual cluster
                for i in range(num_points):
                    if not added[i]:
                        clusters.append([i])
                break
            i, j = np.unravel_index(np.argmax(sim_copy), sim_copy.shape)
            clusters.append([int(i), int(j)])
            added[i] = True
            added[j] = True
            sim_copy[i, :] = 0
            sim_copy[:, i] = 0
            sim_copy[j, :] = 0
            sim_copy[:, j] = 0

    return clusters



import asyncio
import io
import json
import logging
import os
from os.path import dirname, exists, join
import shutil
import subprocess
import tempfile
import traceback
from itertools import product
from shutil import which
from time import sleep, time
from typing import Any

import json_repair
import Levenshtein
from html2image import Html2Image
from pdf2image import convert_from_path
from PIL import Image as PILImage
from pptx.dml.color import RGBColor
from pptx.oxml import parse_xml
from pptx.parts.image import Image
from pptx.shapes.group import GroupShape
from pptx.text.text import _Paragraph, _Run
from pptx.util import Length, Pt
from pydantic import BaseModel
from tenacity import RetryCallState, retry, stop_after_attempt, wait_fixed


class Language(BaseModel):
    lid: str

    # mark the character set
    @property
    def latin(self) -> bool:
        return not self.cjk

    @property
    def cjk(self) -> bool:
        return self.lid in ["zh", "ja", "ko"]

    @classmethod
    def chinese(cls):
        return Language(lid="zh")

    @classmethod
    def english(cls):
        return Language(lid="en")


def get_logger(name="pptagent", level=None):
    """
    Get a logger with the specified name and level.

    Args:
        name (str): The name of the logger.
        level (int): The logging level (default: logging.INFO).

    Returns:
        logging.Logger: A configured logger instance.
    """
    if level is None:
        level = int(os.environ.get("LOG_LEVEL", logging.INFO))

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Check if the logger already has handlers to avoid duplicates
    if not logger.handlers:
        # Create console handler and set level
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)

        # Create formatter
        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s"
        )
        console_handler.setFormatter(formatter)

        # Add handler to logger
        logger.addHandler(console_handler)

    return logger


logger = get_logger(__name__)

if which("soffice") is None:
    logging.warning("soffice is not installed, pptx to images conversion will not work")

# Set of supported image extensions
IMAGE_EXTENSIONS: set[str] = {
    "bmp",
    "jpg",
    "jpeg",
    "pgm",
    "png",
    "ppm",
    "tif",
    "tiff",
    "webp",
}

# Common colors and measurements
BLACK = RGBColor(0, 0, 0)
YELLOW = RGBColor(255, 255, 0)
BLUE = RGBColor(0, 0, 255)
BORDER_LEN = Pt(2)
BORDER_OFFSET = Pt(2)
LABEL_LEN = Pt(24)
FONT_LEN = Pt(20)


def is_image_path(file: str) -> bool:
    """
    Check if a file path is an image based on its extension.

    Args:
        file (str): The file path to check.

    Returns:
        bool: True if the file is an image, False otherwise.
    """
    return file.split(".")[-1].lower() in IMAGE_EXTENSIONS


def runs_merge(paragraph: _Paragraph) -> _Run | None:
    """
    Merge all runs in a paragraph into a single run.

    Args:
        paragraph (_Paragraph): The paragraph to merge runs in.

    Returns:
        Optional[_Run]: The merged run, or None if there are no runs.
    """
    runs = paragraph.runs

    # Handle field codes
    if len(runs) == 0:
        runs = [
            _Run(r, paragraph)
            for r in parse_xml(paragraph._element.xml.replace("fld", "r")).r_lst
        ]
    if len(runs) == 1:
        return runs[0]
    if len(runs) == 0:
        return None

    # Find the run with the most text
    run = max(runs, key=lambda x: len(x.text))
    run.text = paragraph.text

    # Remove other runs
    for r in runs:
        if r != run:
            r._r.getparent().remove(r._r)
    return run


def older_than(filepath: str, seconds: int = 10, wait: bool = False) -> bool:
    """
    Check if a file is older than a specified number of seconds.

    Args:
        filepath (str): The path to the file.
        seconds (int): The number of seconds to check against.
        wait (bool): Whether to wait for the file to exist.

    Returns:
        bool: True if the file is older than the specified number of seconds, False otherwise.
    """
    if not os.path.exists(filepath):
        while wait:
            logger.info("waiting for: %s", filepath)
            sleep(1)
            if os.path.exists(filepath):
                sleep(seconds)
                return True
        return False
    file_creation_time = os.path.getctime(filepath)
    current_time = time()
    return seconds < (current_time - file_creation_time)


def edit_distance(text1: str, text2: str) -> float:
    """
    Calculate the normalized edit distance between two strings.

    Args:
        text1 (str): The first string.
        text2 (str): The second string.

    Returns:
        float: The normalized edit distance (0.0 to 1.0, where 1.0 means identical).
    """
    if not text1 and not text2:
        return 1.0
    return 1 - Levenshtein.distance(text1, text2) / max(len(text1), len(text2))


def tenacity_log(retry_state: RetryCallState) -> None:
    """
    Log function for tenacity retries.

    Args:
        retry_state (RetryCallState): The retry state.
    """
    logger.warning("tenacity retry: %s", retry_state)
    traceback.print_tb(retry_state.outcome.exception().__traceback__)


def get_json_from_response(response: str) -> dict[str, Any]:
    """
    Extract JSON from a text response.

    Args:
        response (str): The response text.

    Returns:
        Dict[str, Any]: The extracted JSON.

    Raises:
        Exception: If JSON cannot be extracted from the response.
    """
    response = response.strip()

    try:
        return json.loads(response)
    except Exception:
        pass

    # Try to extract JSON from markdown code blocks
    l, r = response.rfind("```json"), response.rfind("```")
    if l != -1 and r != -1:
        json_obj = json_repair.loads(response[l + 7 : r].strip())
        if isinstance(json_obj, (dict, list)):
            return json_obj

    # Try to find JSON by looking for matching braces
    open_braces = []
    close_braces = []

    for i, char in enumerate(response):
        if char == "{" or char == "[":
            open_braces.append(i)
        elif char == "}" or char == "]":
            close_braces.append(i)

    for i, j in product(open_braces, reversed(close_braces)):
        if i > j:
            continue
        try:
            json_obj = json_repair.loads(response[i : j + 1])
            if isinstance(json_obj, (dict, list)):
                return json_obj
        except Exception:
            pass

    raise Exception("JSON not found in the given output", response)


# Create a tenacity decorator with custom settings
def tenacity_decorator(_func=None, *, wait: int = 3, stop: int = 5):
    def decorator(func):
        return retry(wait=wait_fixed(wait), stop=stop_after_attempt(stop))(func)

    if _func is None:
        # Called with arguments
        return decorator
    else:
        # Called without arguments
        return decorator(_func)


TABLE_CSS = """
table {
    border-collapse: collapse;  /* Merge borders */
    width: auto;               /* Width adapts to content */
    font-family: SimHei, Arial, sans-serif;  /* Font supporting Chinese characters */
    background: white;
}
th, td {
    border: 1px solid black;  /* Add borders */
    padding: 8px;             /* Cell padding */
    text-align: center;       /* Center text */
}
th {
    background-color: #f2f2f2; /* Header background color */
}
"""


def manual_scan_crop(img_path: str):
    """Manually detect and crop image boundaries by scanning pixels."""
    img = PILImage.open(img_path).convert("RGB")
    width, height = img.size
    pixels = img.load()
    left, top, right, bottom = width, height, 0, 0
    found_content = False

    # Scan all pixels to find non-white areas
    for y in range(height):
        for x in range(width):
            r, g, b = pixels[x, y]
            # Detect non-white pixels (using a relaxed threshold to account for anti-aliasing)
            if r < 248 or g < 248 or b < 248:
                left = min(left, x)
                right = max(right, x)
                top = min(top, y)
                bottom = max(bottom, y)
                found_content = True

    if found_content:
        padding = 20
        bbox = (
            max(0, left - padding),
            max(0, top - padding),
            min(width, right + 1 + padding),
            min(height, bottom + 1 + padding),
        )

        cropped_img = img.crop(bbox)
        cropped_img.save(img_path)


def get_html_table_image(html: str, output_path: str, css: str = None):
    """
    Convert a html table to the image

    Args:
    html (str): html text containing a table
    output_path (str): Output image path, defaults to 'table_cropped.png'

    Returns:
    str: The path of the generated image
    """
    if css is None:
        css = TABLE_CSS
    parent_dir, base_name = os.path.split(output_path)

    if parent_dir and not os.path.exists(parent_dir):
        os.makedirs(parent_dir)

    hti = Html2Image(
        disable_logging=True,
        output_path=parent_dir if parent_dir else ".",
        custom_flags=["--no-sandbox", "--headless", "--disable-gpu"],
    )
    hti.browser.use_new_headless = None
    hti.screenshot(
        html_str=html,
        css_str=css,
        save_as=base_name,
        size=(1000, 600),
    )
    manual_scan_crop(output_path)


@tenacity_decorator
def ppt_to_images(file: str, output_dir: str):
    assert exists(file), f"File {file} does not exist"
    if exists(output_dir):
        logger.warning(f"ppt2images: {output_dir} already exists")
    os.makedirs(output_dir, exist_ok=True)
    with tempfile.TemporaryDirectory() as temp_dir:
        command_list = [
            "soffice",
            "--headless",
            "--convert-to",
            "pdf",
            file,
            "--outdir",
            temp_dir,
        ]
        process = subprocess.Popen(
            command_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = process.communicate()
        if process.returncode != 0:
            raise RuntimeError(f"soffice failed with error: {err.decode()}")

        for f in os.listdir(temp_dir):
            if not f.endswith(".pdf"):
                continue
            temp_pdf = join(temp_dir, f)
            images = convert_from_path(temp_pdf, dpi=72)
            for i, img in enumerate(images):
                img.save(join(output_dir, f"slide_{i + 1:04d}.jpg"))
            return

        raise RuntimeError(
            f"No PDF file was created in the temporary directory: {file}\n"
            f"Output: {out.decode()}\n"
            f"Error: {err.decode()}"
        )


@tenacity_decorator
async def ppt_to_images_async(file: str, output_dir: str):
    assert exists(file), f"File {file} does not exist"
    if exists(output_dir):
        logger.debug(f"ppt2images: {output_dir} already exists")
    os.makedirs(output_dir, exist_ok=True)

    with tempfile.TemporaryDirectory() as temp_dir:
        command_list = [
            "soffice",
            "--headless",
            "--convert-to",
            "pdf",
            file,
            "--outdir",
            temp_dir,
        ]

        process = await asyncio.create_subprocess_exec(
            *command_list,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()
        if process.returncode != 0:
            raise RuntimeError(f"soffice failed with error: {stderr.decode()}")
        for f in os.listdir(temp_dir):
            if not f.endswith(".pdf"):
                continue
            temp_pdf = join(temp_dir, f)
            images = convert_from_path(temp_pdf, dpi=72)
            for i, img in enumerate(images):
                img.save(join(output_dir, f"slide_{i + 1:04d}.jpg"))
            return

        raise RuntimeError(
            f"No PDF file was created in the temporary directory: {file}\n"
            f"Output: {stdout.decode()}\n"
            f"Error: {stderr.decode()}"
        )


def parsing_image(image: Image, image_path: str) -> str:
    # Handle WMF images (PDFs)
    if image.ext == "wmf":
        image_path = image_path.replace(".wmf", ".png")
        if not exists(image_path):
            wmf_to_images(image.blob, image_path)
    # Check for supported image types
    elif image.ext in ["webp", "tiff"]:
        image_path = image_path.replace(".webp", ".png").replace(".tiff", ".png")
        pil_image = PILImage.open(io.BytesIO(image.blob))
        if not exists(image_path):
            pil_image.save(image_path, "PNG")
        return image_path
    elif image.ext not in IMAGE_EXTENSIONS:
        raise ValueError(f"Unsupported image type {image.ext}")

    # Save image if it doesn't exist
    if not exists(image_path):
        with open(image_path, "wb") as f:
            f.write(image.blob)
    return image_path


@tenacity_decorator
def wmf_to_images(blob: bytes, filepath: str):
    if not filepath.endswith(".png"):
        raise ValueError("filepath must end with .png")
    dirname = os.path.dirname(filepath)
    base_name = os.path.basename(filepath).removesuffix(".png")
    with tempfile.TemporaryDirectory() as temp_dir:
        with open(join(temp_dir, f"{base_name}.wmf"), "wb") as f:
            f.write(blob)
        command_list = [
            "soffice",
            "--headless",
            "--convert-to",
            "png",
            join(temp_dir, f"{base_name}.wmf"),
            "--outdir",
            dirname,
        ]
        subprocess.run(command_list, check=True, stdout=subprocess.DEVNULL)

    assert exists(filepath), f"File {filepath} does not exist"


def parse_groupshape(groupshape: GroupShape) -> list[dict[str, Length]]:
    """
    Parse a group shape to get the bounds of its child shapes.

    Args:
        groupshape (GroupShape): The group shape to parse.

    Returns:
        List[Dict[str, Length]]: The bounds of the child shapes.

    Raises:
        AssertionError: If the input is not a GroupShape.
    """
    assert isinstance(groupshape, GroupShape), "Input must be a GroupShape"

    # Get group bounds
    group_top_left_x = groupshape.left
    group_top_left_y = groupshape.top
    group_width = groupshape.width
    group_height = groupshape.height

    # Get shape bounds
    shape_top_left_x = min([sp.left for sp in groupshape.shapes])
    shape_top_left_y = min([sp.top for sp in groupshape.shapes])
    shape_width = (
        max([sp.left + sp.width for sp in groupshape.shapes]) - shape_top_left_x
    )
    shape_height = (
        max([sp.top + sp.height for sp in groupshape.shapes]) - shape_top_left_y
    )

    # Calculate bounds for each shape in the group
    group_shape_xy = []
    for sp in groupshape.shapes:
        group_shape_left = (
            sp.left - shape_top_left_x
        ) * group_width / shape_width + group_top_left_x
        group_shape_top = (
            sp.top - shape_top_left_y
        ) * group_height / shape_height + group_top_left_y
        group_shape_width = sp.width * group_width / shape_width
        group_shape_height = sp.height * group_height / shape_height

        group_shape_xy.append(
            {
                "left": Length(group_shape_left),
                "top": Length(group_shape_top),
                "width": Length(group_shape_width),
                "height": Length(group_shape_height),
            }
        )

    return group_shape_xy


def is_primitive(obj: Any) -> bool:
    """
    Check if an object is a primitive type or a collection of primitive types.

    Args:
        obj (Any): The object to check.

    Returns:
        bool: True if the object is a primitive type or a collection of primitive types, False otherwise.
    """
    if isinstance(obj, (list, tuple, set, frozenset)):
        return all(is_primitive(item) for item in obj)

    return isinstance(
        obj, (int, float, complex, bool, str, bytes, bytearray, type(None))
    )


DEFAULT_EXCLUDE: set[str] = {"element", "language_id", "ln", "placeholder_format"}


def dict_to_object(
    dict_obj: dict[str, Any], obj: Any, exclude: set[str] | None = None
) -> None:
    """
    Apply dictionary values to an object.

    Args:
        dict_obj (Dict[str, Any]): The dictionary with values to apply.
        obj (Any): The object to apply values to.
        exclude (Optional[Set[str]]): The keys to exclude.
    """
    if exclude is None:
        exclude = set()

    for key, value in dict_obj.items():
        if key not in exclude and value is not None:
            setattr(obj, key, value)


def package_join(*paths: str) -> str:
    """
    Join paths with the appropriate separator for the platform.

    Args:
        *paths: The paths to join.

    Returns:
        str: The joined path.
    """
    _dir = dirname(__file__)
    return join(_dir, *paths)


class Config:
    """
    Configuration class for the application.
    """

    def __init__(
        self,
        rundir: str | None = None,
        session_id: str | None = None,
    ):
        """
        Initialize the configuration.

        Args:
            rundir (Optional[str]): The run directory.
            session_id (Optional[str]): The session ID.
            debug (bool): Whether to enable debug mode.
        """
        if rundir is not None:
            self.set_rundir(rundir)
        elif session_id is not None:
            self.set_session(session_id)
        else:
            raise ValueError("No session ID or run directory provided")

    def set_session(self, session_id: str) -> None:
        """
        Set the session ID and update the run directory.

        Args:
            session_id (str): The session ID.
        """
        self.session_id = session_id
        self.set_rundir(f"./runs/{session_id}")

    def set_rundir(self, rundir: str) -> None:
        """
        Set the run directory and create necessary subdirectories.

        Args:
            rundir (str): The run directory.
        """
        self.RUN_DIR = rundir
        self.IMAGE_DIR = join(self.RUN_DIR, "images")

        for the_dir in [self.RUN_DIR, self.IMAGE_DIR]:
            os.makedirs(the_dir, exist_ok=True)

    def set_debug(self, debug: bool) -> None:
        """
        Set the debug mode.

        Args:
            debug (bool): Whether to enable debug mode.
        """
        self.DEBUG = debug

    def remove_rundir(self) -> None:
        """
        Remove the run directory and its subdirectories.
        """
        if exists(self.RUN_DIR):
            shutil.rmtree(self.RUN_DIR)
        if exists(self.IMAGE_DIR):
            shutil.rmtree(self.IMAGE_DIR)

    def __repr__(self) -> str:
        """
        Get a string representation of the configuration.

        Returns:
            str: A string representation of the configuration.
        """
        attrs = []
        for attr in dir(self):
            if not attr.startswith("_") and not callable(getattr(self, attr)):
                attrs.append(f"{attr}={getattr(self, attr)}")
        return f"Config({', '.join(attrs)})"



import tempfile
from os.path import join, exists
from test.conftest import test_config

import pytest

from pptagent.model_utils import parse_pdf


@pytest.mark.parse
@pytest.mark.asyncio
async def test_parse_pdf():
    with tempfile.TemporaryDirectory() as temp_dir:
        await parse_pdf(
            join(test_config.document, "source.pdf"),
            temp_dir,
        )
        assert exists(join(temp_dir, "source.md"))



import asyncio
import os
from collections import defaultdict
from functools import partial

from aiometer import run_all
from jinja2 import Template

from pptagent.agent import Agent
from pptagent.llms import AsyncLLM
from pptagent.model_utils import (
    get_cluster,
    get_image_embedding,
    images_cosine_similarity,
    language_id,
)
from pptagent.presentation import Picture, Presentation, SlidePage
from pptagent.response import SlideSchema
from os.path import join
from pptagent.utils import (
    Config,
    get_logger,
    is_image_path,
    package_join,
)

logger = get_logger(__name__)

CATEGORY_SPLIT_TEMPLATE = Template(
    open(package_join("prompts", "category_split.txt"), encoding="utf-8").read()
)
ASK_CATEGORY_PROMPT = open(
    package_join("prompts", "ask_category.txt"), encoding="utf-8"
).read()


class SlideInducter:
    """
    Stage I: Presentation Analysis.
    This stage is to analyze the presentation: cluster slides into different layouts, and extract content schema for each layout.
    """

    def __init__(
        self,
        prs: Presentation,
        ppt_image_folder: str,
        template_image_folder: str,
        config: Config,
        image_models: list,
        language_model: AsyncLLM,
        vision_model: AsyncLLM,
        use_assert: bool = True,
    ):
        """
        Initialize the SlideInducter.

        Args:
            prs (Presentation): The presentation object.
            ppt_image_folder (str): The folder containing PPT images.
            template_image_folder (str): The folder containing normalized slide images.
            config (Config): The configuration object.
            image_models (list): A list of image models.
        """
        self.prs = prs
        self.config = config
        self.ppt_image_folder = ppt_image_folder
        self.template_image_folder = template_image_folder
        self.language_model = language_model
        self.vision_model = vision_model
        self.image_models = image_models
        self.schema_extractor = Agent(
            "schema_extractor",
            {
                "language": language_model,
            },
        )
        if not use_assert:
            return

        num_template_images = sum(
            is_image_path(f) for f in os.listdir(template_image_folder)
        )
        num_ppt_images = sum(is_image_path(f) for f in os.listdir(ppt_image_folder))
        num_slides = len(prs.slides)

        if not (num_template_images == num_ppt_images == num_slides):
            raise ValueError(
                f"Slide count mismatch detected:\n"
                f"- Presentation slides: {num_slides}\n"
                f"- Template images: {num_template_images} ({template_image_folder})\n"
                f"- PPT images: {num_ppt_images} ({ppt_image_folder})\n"
                f"All counts must be equal."
            )

    async def category_split(self):
        """
        Async version: Split slides into categories based on their functional purpose.
        """
        functional_cluster = await self.language_model(
            CATEGORY_SPLIT_TEMPLATE.render(slides=self.prs.to_text()),
            return_json=True,
        )
        assert isinstance(functional_cluster, dict) and all(
            isinstance(k, str) and isinstance(v, list)
            for k, v in functional_cluster.items()
        ), "Functional cluster must be a dictionary with string keys and list values"
        functional_slides = set(sum(functional_cluster.values(), []))
        content_slides_index = set(range(1, len(self.prs) + 1)) - functional_slides

        return content_slides_index, functional_cluster

    async def layout_split(
        self, content_slides_index: set[int], layout_induction: dict
    ):
        """
        Async version: Cluster slides into different layouts.
        """
        embeddings = get_image_embedding(self.template_image_folder, *self.image_models)
        assert len(embeddings) == len(self.prs)
        content_split = defaultdict(list)
        for slide_idx in content_slides_index:
            slide = self.prs.slides[slide_idx - 1]
            content_type = slide.get_content_type()
            layout_name = slide.slide_layout_name
            content_split[(layout_name, content_type)].append(slide_idx)

        async with asyncio.TaskGroup() as tg:
            for (layout_name, content_type), slides in content_split.items():
                sub_embeddings = [
                    embeddings[f"slide_{slide_idx:04d}.jpg"] for slide_idx in slides
                ]
                similarity = images_cosine_similarity(sub_embeddings)
                for cluster in get_cluster(similarity):
                    slide_indexs = [slides[i] for i in cluster]
                    template_id = max(
                        slide_indexs,
                        key=lambda x: len(self.prs.slides[x - 1].shapes),
                    )

                    tg.create_task(
                        self.vision_model(
                            ASK_CATEGORY_PROMPT,
                            join(self.ppt_image_folder, f"slide_{template_id:04d}.jpg"),
                        )
                    ).add_done_callback(
                        lambda f,
                        tid=template_id,
                        sidxs=slide_indexs,
                        ctype=content_type: layout_induction[
                            f.result() + ":" + ctype
                        ].update({"template_id": tid, "slides": sidxs})
                    )

    async def layout_induct(self):
        """
        Async version: Perform layout induction for the presentation.
        """
        layout_induction = defaultdict(lambda: defaultdict(list))
        content_slides_index, functional_cluster = await self.category_split()
        for layout_name, cluster in functional_cluster.items():
            layout_induction[layout_name]["slides"] = cluster
            layout_induction[layout_name]["template_id"] = cluster[0]

        functional_keys = list(layout_induction.keys())
        function_slides_index = set()
        for layout_name, cluster in layout_induction.items():
            function_slides_index.update(cluster["slides"])
        used_slides_index = function_slides_index.union(content_slides_index)
        for i in range(len(self.prs.slides)):
            if i + 1 not in used_slides_index:
                content_slides_index.add(i + 1)
        await self.layout_split(content_slides_index, layout_induction)
        layout_induction["functional_keys"] = functional_keys
        return layout_induction

    async def content_induct(
        self,
        layout_induction: dict,
        max_at_once: int | None = None,
        max_per_second: int | None = None,
    ):
        """
        Async version: Perform content schema extraction for the presentation.
        """
        partial_funcs = []
        for layout_name, cluster in layout_induction.items():
            if layout_name == "functional_keys" or "content_schema" in cluster:
                continue
            slide: SlidePage = self.prs.slides[cluster["template_id"] - 1]

            contents = [para.text for para in slide.iter_paragraphs()] + [
                shape.caption for shape in slide.shape_filter(Picture)
            ]
            partial_funcs.append(
                partial(
                    self.schema_extractor,
                    slide=slide.to_html(),
                    response_format=SlideSchema.response_model(contents),
                    slide_idx=cluster["template_id"],
                )
            )

        schemas = await run_all(
            partial_funcs, max_at_once=max_at_once, max_per_second=max_per_second
        )
        for layout_name, (_, schema) in zip(layout_induction.keys(), schemas):
            layout_induction[layout_name].update(schema)

        layout_induction["language"] = language_id(slide.to_text()).model_dump()
        return layout_induction



from contextvars import ContextVar
from typing import Literal

from pydantic import BaseModel

from pptagent.utils import edit_distance

_allowed_contents = ContextVar("allowed_contents")


class SlideElement(BaseModel):
    name: str
    data: list[str]
    type: Literal["text", "image"]

    def model_post_init(self, _):
        self.data = [
            max(_allowed_contents.get(), key=lambda x: edit_distance(x, d))
            for d in self.data
        ]


class SlideSchema(BaseModel):
    elements: list[SlideElement]

    @classmethod
    def response_model(cls, content_fields: list[str]) -> type[BaseModel]:
        _allowed_contents.set(content_fields)
        return cls



from os.path import join as pjoin
from pathlib import Path
from test.conftest import test_config

import pytest

from pptagent.induct import SlideInducter
from pptagent.multimodal import ImageLabler
from pptagent.presentation import Presentation

CUTOFF = 8


def prepare_slides(prs: Presentation):
    slides = []
    slide_image_folder = pjoin(test_config.template, "slide_images")
    template_image_folder = pjoin(test_config.template, "template_images")
    for i, sld in enumerate(prs.slides):
        if i < CUTOFF:
            slides.append(sld)
        else:
            Path(pjoin(slide_image_folder, f"slide_{i + 1:04d}.jpg")).unlink(
                missing_ok=True
            )
            Path(pjoin(template_image_folder, f"slide_{i + 1:04d}.jpg")).unlink(
                missing_ok=True
            )
    prs.slides = slides
    return prs


@pytest.mark.asyncio
@pytest.mark.llm
async def test_layout_induct():
    prs = Presentation.from_file(
        pjoin(test_config.template, "source.pptx"), test_config.config
    )
    labler = ImageLabler(prs, test_config.config)
    labler.apply_stats(test_config.get_image_stats())
    prs = prepare_slides(prs)

    inducter = SlideInducter(
        prs,
        pjoin(test_config.template, "slide_images"),
        pjoin(test_config.template, "template_images"),
        test_config.config,
        test_config.image_model,
        test_config.language_model,
        test_config.vision_model,
    )
    await inducter.layout_induct()


@pytest.mark.asyncio
@pytest.mark.llm
async def test_content_induct():
    prs = Presentation.from_file(
        pjoin(test_config.template, "source.pptx"), test_config.config
    )
    labler = ImageLabler(prs, test_config.config)
    labler.apply_stats(test_config.get_image_stats())
    prs = prepare_slides(prs)

    inducter = SlideInducter(
        prs,
        pjoin(test_config.template, "slide_images"),
        pjoin(test_config.template, "template_images"),
        test_config.config,
        test_config.image_model,
        test_config.language_model,
        test_config.vision_model,
    )
    layout_induction = {}
    for layout_name, cluster in test_config.get_slide_induction().items():
        cluster.pop("elements")
        layout_induction[layout_name] = cluster
        break
    await inducter.content_induct(layout_induction=layout_induction)



from .induct import SlideSchema
from .outline import Outline, OutlineItem
from .pptgen import EditorOutput, LayoutChoice

__all__ = [
    "LogicHeadings",
    "EditorOutput",
    "LayoutChoice",
    "SlideSchema",
    "Outline",
    "OutlineItem",
]



import inspect
import re
import traceback
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from functools import partial
from typing import Any

from bs4 import BeautifulSoup
from mistune import HTMLRenderer, create_markdown
from PIL import Image
from pptx.enum.text import PP_ALIGN
from pptx.oxml import parse_xml
from pptx.shapes.base import BaseShape
from pptx.shapes.graphfrm import GraphicFrame as PPTXGraphicFrame
from pptx.text.text import _Run
from pptx.util import Pt

from pptagent.document import Document
from pptagent.presentation import Closure, ClosureType, Picture, ShapeElement, SlidePage
from pptagent.utils import get_logger, runs_merge

logger = get_logger(__name__)
TABLE_REGEX = re.compile(r".*table_[0-9a-fA-F]{4}\.png$")


class SlideRenderer(HTMLRenderer):
    """
    A renderer that does not render lists.
    """

    def list(self, text: str, ordered: bool, **attrs: Any) -> str:
        return text

    def list_item(self, text: str) -> str:
        return text


markdown = create_markdown(renderer=SlideRenderer(), plugins=["strikethrough"])


class SlideEditError(Exception):
    """
    Exception raised when an edit operation fails.
    """


@dataclass
class HistoryMark:
    """
    Mark the execution status of the API call, comment and a line of code.
    """

    API_CALL_ERROR = "api_call_error"
    API_CALL_CORRECT = "api_call_correct"
    COMMENT_CORRECT = "comment_correct"
    COMMENT_ERROR = "comment_error"
    CODE_RUN_ERROR = "code_run_error"
    CODE_RUN_CORRECT = "code_run_correct"


class CodeExecutor:
    """
    Execute code actions and manage API call history, and providing error feedback.
    """

    def __init__(self, retry_times: int):
        """
        Initialize the CodeExecutor.

        Args:
            retry_times (int): The number of times to retry failed actions.
        """
        self.api_history = []
        self.command_history = []
        self.code_history = []
        self.retry_times = retry_times
        self.registered_functions = API_TYPES.all_funcs()
        self.function_regex = re.compile(r"^[a-z]+_[a-z_]+\(.+\)")

    @classmethod
    def get_apis_docs(
        cls,
        funcs: list[callable],
        show_doc: bool = True,
        show_return: bool = True,
        ignore_keys: list[str] | None = None,
    ) -> str:
        """
        Get the documentation for a list of API functions.

        Args:
            funcs (list[callable]): A list of functions to document.
            show_example (bool): Whether to show examples in the documentation.

        Returns:
            str: The formatted API documentation.
        """
        if ignore_keys is None:
            ignore_keys = {"slide", "self", "doc"}
        api_doc = []
        for func in funcs:
            sig = inspect.signature(func)
            params = []
            for name, param in sig.parameters.items():
                if name in ignore_keys:
                    continue
                param_str = name
                if param.annotation != inspect.Parameter.empty:
                    param_str += f": {param.annotation.__name__}"
                if param.default != inspect.Parameter.empty:
                    param_str += f" = {repr(param.default)}"
                params.append(param_str)
            signature = f"def {func.__name__}({', '.join(params)})"
            if show_return and sig.return_annotation != inspect.Parameter.empty:
                signature += f" -> {sig.return_annotation.__name__}"
            if show_doc and inspect.getdoc(func) is not None:
                doc = "\t" + inspect.getdoc(func)
            else:
                doc = ""
            signature += f"\n{doc}"
            api_doc.append(signature)
        return "\n".join(api_doc)

    def execute_actions(
        self,
        actions: str,
        edit_slide: SlidePage,
        doc: Document,
        found_code: bool = False,
    ) -> tuple[str, str] | None:
        """
        Execute a series of actions on a slide.

        Args:
            actions (str): The actions to execute.
            edit_slide (SlidePage): The slide to edit.
            found_code (bool): Whether code was found in the actions.

        Returns:
            tuple: The API lines and traceback if an error occurs.
            None: If no error occurs.
        """
        api_calls = actions.strip().split("\n")
        self.api_history.append(
            [HistoryMark.API_CALL_ERROR, edit_slide.slide_idx, actions]
        )
        for line_idx, line in enumerate(api_calls):
            try:
                if line_idx == len(api_calls) - 1 and not found_code:
                    raise SlideEditError(
                        "No code block found in the output, please output the api calls without any prefix."
                    )
                if line.startswith("def"):
                    raise SlideEditError("The function definition were not allowed.")
                if line.startswith("#"):
                    if len(self.command_history) != 0:
                        self.command_history[-1][0] = HistoryMark.COMMENT_CORRECT
                    self.command_history.append([HistoryMark.COMMENT_ERROR, line, None])
                    continue
                if not self.function_regex.match(line):
                    continue
                found_code = True
                func = line.split("(")[0]
                if func not in self.registered_functions:
                    raise SlideEditError(f"The function {func} is not defined.")
                # only one of clone and del can be used in a row
                if func.startswith("clone") or func.startswith("del"):
                    tag = func.split("_")[0]
                    if (
                        self.command_history[-1][-1] is None
                        or self.command_history[-1][-1] == tag
                    ):
                        self.command_history[-1][-1] = tag
                    else:
                        raise SlideEditError(
                            "Invalid command: Both 'clone_paragraph' and 'del_paragraph'/'del_image' are used within a single command. "
                            "Each command must only perform one of these operations based on the quantity_change."
                        )
                self.code_history.append([HistoryMark.CODE_RUN_ERROR, line, None])
                partial_func = partial(self.registered_functions[func], edit_slide)
                if func == "replace_image":
                    partial_func = partial(partial_func, doc)
                eval(line, {}, {func: partial_func})
                self.code_history[-1][0] = HistoryMark.CODE_RUN_CORRECT
            except Exception as e:
                if not isinstance(e, SlideEditError):
                    logger.warning(f"Encountered unknown error: {e}")

                trace_msg = traceback.format_exc()
                if len(self.code_history) != 0:
                    self.code_history[-1][-1] = trace_msg
                api_lines = (
                    "\n".join(api_calls[: line_idx - 1])
                    + f"\n--> Error Line: {line}\n"
                    + "\n".join(api_calls[line_idx + 1 :])
                )
                return api_lines, trace_msg
        if len(self.command_history) != 0:
            self.command_history[-1][0] = HistoryMark.COMMENT_CORRECT
        self.api_history[-1][0] = HistoryMark.API_CALL_CORRECT

    def __add__(self, other):
        self.api_history.append(other.api_history)
        self.command_history.extend(other.command_history)
        self.code_history.extend(other.code_history)
        return self


# supporting functions
def element_index(slide: SlidePage, element_id: int) -> ShapeElement:
    """
    Find the an element in a slide.

    Args:
        slide (SlidePage): The slide
        element_id (int): The ID of the element.

    Returns:
        ShapeElement: The shape corresponding to the element ID.
    """
    for shape in slide:
        if shape.shape_idx == element_id:
            return shape
    raise SlideEditError(
        f"Cannot find element {element_id}, is it deleted or not exist?"
    )


@dataclass
class TextBlock:
    text: str
    bold: bool = False
    italic: bool = False
    code: bool = False
    strikethrough: bool = False
    href: str = None

    def build_run(self, run: _Run):
        if self.bold:
            run.font.bold = True
        if self.italic:
            run.font.italic = True
        if self.code:
            run.font.name = "Consolas"
        if self.strikethrough:
            run.font.strikethrough = True
        if self.href is not None:
            run.hyperlink.address = self.href

        run.text = self.text


MARKDOWN_STYLES = {
    "strong": "bold",
    "em": "italic",
    "code": "code",
    "del": "strikethrough",
}


def process_element(element, styles=None) -> list[TextBlock]:
    if styles is None:
        styles = {}

    result = []

    if isinstance(element, str):
        result.append(TextBlock(element, **styles))
    else:
        if element.name == "a":
            new_styles = styles.copy()
            for child in element.children:
                blocks = process_element(child, new_styles)
                for block in blocks:
                    block.href = element.get("href")
                result.extend(blocks)
        elif MARKDOWN_STYLES.get(element.name):
            new_styles = styles.copy()
            new_styles[MARKDOWN_STYLES[element.name]] = True
            for child in element.children:
                result.extend(process_element(child, new_styles))
        else:
            for child in element.children:
                result.extend(process_element(child, styles))

    return result


def replace_para(paragraph_id: int, new_text: str, shape: BaseShape):
    """
    Replace the text of a paragraph in a shape.
    """
    para = shape.text_frame.paragraphs[paragraph_id]
    html = markdown(new_text).strip()
    soup = BeautifulSoup(html, "html.parser")
    blocks = process_element(soup)

    empty_run = runs_merge(para)
    empty_run.text = ""
    for _ in range(len(blocks) - 1):
        empty_run._r.addnext(parse_xml(empty_run._r.xml))
    for block, run in zip(blocks, para.runs):
        block.build_run(run)


def clone_para(paragraph_id: int, shape: BaseShape):
    """
    Clone a paragraph in a shape.
    """
    para = shape.text_frame.paragraphs[paragraph_id]
    shape.text_frame.paragraphs[-1]._element.addnext(parse_xml(para._element.xml))


def del_para(paragraph_id: int, shape: BaseShape):
    """
    Delete a paragraph from a shape.
    """
    tf = shape.text_frame
    # Textframe need to have at least one paragraph
    if len(tf.paragraphs) == 1:
        tf.text = ""
        return
    para = tf.paragraphs[paragraph_id]
    para._element.getparent().remove(para._element)


def add_table(table_data: list[list[str]], table: PPTXGraphicFrame):
    rows = len(table_data)
    cols = len(table_data[0])

    max_lengths = [max(len(row[j]) for row in table_data) for j in range(cols)]
    total_length = sum(max_lengths)
    for j in range(cols):
        col_width = int((max_lengths[j] / total_length) * table.width)
        table.table.columns[j].width = col_width

    for i in range(rows):
        for j in range(cols):
            table.table.cell(i, j).text = table_data[i][j]


def merge_cells(merge_area: list[tuple[int, int, int, int]], table: PPTXGraphicFrame):
    if merge_area is None or len(merge_area) == 0:
        return
    for x1, y1, x2, y2 in merge_area:
        table.table.cell(x1, y1).merge(table.table.cell(x2, y2))
        for x, y in zip(range(x1, x2 + 1), range(y1, y2 + 1)):
            tf = table.table.cell(x, y).text_frame
            for p in tf.paragraphs:
                p.alignment = PP_ALIGN.CENTER


# api functions
def del_paragraph(slide: SlidePage, div_id: int, paragraph_id: int):
    """
    Delete a paragraph from a slide.

    Args:
        slide (SlidePage): The slide containing the paragraph.
        div_id (int): The ID of the division containing the paragraph.
        paragraph_id (int): The ID of the paragraph to delete.

    """
    shape = element_index(slide, div_id)
    if not shape.text_frame.is_textframe:
        raise SlideEditError(
            f"The element {shape.shape_idx} of slide {slide.slide_idx} does not have a text frame, please check the element id and type of element."
        )
    for para in shape.text_frame.paragraphs:
        if para.idx == paragraph_id:
            para.edited = True
            shape.text_frame.paragraphs.remove(para)
            shape._closures[ClosureType.DELETE].append(
                Closure(partial(del_para, para.real_idx), para.real_idx)
            )
            return
    else:
        raise SlideEditError(
            f"Cannot find the paragraph {paragraph_id} of the element {div_id},"
            "may refer to a non-existed paragraph or you haven't cloned enough paragraphs beforehand."
        )


def del_image(slide: SlidePage, figure_id: int):
    """
    Delete an image from a slide.

    Args:
        slide (SlidePage): The slide containing the image.
        figure_id (int): The ID of the image to delete.
    """
    shape = element_index(slide, figure_id)
    if not isinstance(shape, Picture):
        raise SlideEditError(
            f"The element {shape.shape_idx} of slide {slide.slide_idx} is not a Picture."
        )
    slide.shapes.remove(shape)


def replace_paragraph(slide: SlidePage, div_id: int, paragraph_id: int, text: str):
    """
    Replace the text of a paragraph in a slide.

    Args:
        slide (SlidePage): The slide containing the paragraph.
        div_id (int): The ID of the division containing the paragraph.
        paragraph_id (int): The ID of the paragraph to replace.
        text (str): The new text to replace with.

    Raises:
        SlideEditError: If the paragraph is not found.
    """
    shape = element_index(slide, div_id)
    if not shape.text_frame.is_textframe:
        raise SlideEditError(
            f"The element {shape.shape_idx} of slide {slide.slide_idx} does not have a text frame, please check the element id and type of element."
        )
    for para in shape.text_frame.paragraphs:
        if para.idx == paragraph_id:
            para.text = text
            shape._closures[ClosureType.REPLACE].append(
                Closure(
                    partial(replace_para, para.real_idx, text),
                    para.real_idx,
                )
            )
            para.edited = True
            return
    else:
        raise SlideEditError(
            f"Cannot find the paragraph {paragraph_id} of the element {div_id},"
            "Please: "
            "1. check if you refer to a non-existed paragraph."
            "2. check if you already deleted it."
        )


def replace_image(slide: SlidePage, doc: Document | None, img_id: int, image_path: str):
    """
    Replace an image in a slide.

    Args:
        slide (SlidePage): The slide containing the image.
        img_id (int): The ID of the image to replace.
        image_path (str): The path to the new image.

    """
    shape = element_index(slide, img_id)
    if not isinstance(shape, Picture):
        raise SlideEditError(
            f"The element {shape.shape_idx} of slide {slide.slide_idx} is not a Picture."
        )
    if doc is not None:
        shape.caption = doc.find_media(path=image_path).caption
        try:
            if TABLE_REGEX.match(image_path):
                return replace_image_with_table(shape, doc, image_path)
        except Exception as e:
            logger.warning(
                f"Failed to replace image with table element: {e}, fallback to use image directly."
            )

    img_size = Image.open(image_path).size
    r = min(shape.width / img_size[0], shape.height / img_size[1])
    new_width = img_size[0] * r
    new_height = img_size[1] * r
    shape.top = Pt(shape.top + (shape.height - new_height) / 2)
    shape.width = Pt(new_width)
    shape.height = Pt(new_height)
    shape.img_path = image_path


def clone_paragraph(slide: SlidePage, div_id: int, paragraph_id: int):
    """
    Clone a paragraph in a slide.

    Args:
        slide (SlidePage): The slide containing the paragraph.
        div_id (int): The ID of the division containing the paragraph.
        paragraph_id (int): The ID of the paragraph to clone.

    Mention: the cloned paragraph will have a paragraph_id one greater than the current maximum in the parent element.
    """
    shape = element_index(slide, div_id)
    if not shape.text_frame.is_textframe:
        raise SlideEditError(
            f"The element {shape.shape_idx} of slide {slide.slide_idx} does not have a text frame, please check the element id and type of element."
        )
    max_idx = (
        max([para.idx for para in shape.text_frame.paragraphs])
        if shape.text_frame.paragraphs
        else 0
    )
    for para in shape.text_frame.paragraphs:
        if para.idx != paragraph_id:
            continue
        shape.text_frame.paragraphs.append(deepcopy(para))
        shape.text_frame.paragraphs[-1].idx = max_idx + 1
        shape.text_frame.paragraphs[-1].real_idx = len(shape.text_frame.paragraphs) - 1
        shape._closures[ClosureType.CLONE].append(
            Closure(
                partial(clone_para, para.real_idx),
                para.real_idx,
            )
        )
        return
    raise SlideEditError(
        f"Cannot find the paragraph {paragraph_id} of the element {div_id}, may refer to a non-existed paragraph."
    )


def replace_image_with_table(shape: Picture, doc: Document, image_path: str):
    table = doc.find_media(path=image_path)
    shape.is_table = True
    shape.grid = (len(table.cells), len(table.cells[0]))
    shape._closures[ClosureType.REPLACE].append(
        Closure(partial(add_table, table.cells))
    )
    shape._closures[ClosureType.MERGE].append(
        Closure(partial(merge_cells, table.merge_area))
    )
    return


class API_TYPES(Enum):
    Agent = [
        replace_image,
        del_image,
        clone_paragraph,
        replace_paragraph,
        del_paragraph,
    ]

    @classmethod
    def all_funcs(cls) -> dict[str, callable]:
        funcs = {}
        for attr in dir(cls):
            if attr.startswith("__"):
                continue
            funcs |= {func.__name__: func for func in getattr(cls, attr).value}
        return funcs



import base64
import re
import threading
from dataclasses import dataclass
from enum import Enum

import torch
from oaib import Auto
from openai import AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletion
from pydantic import BaseModel

from pptagent.utils import get_json_from_response, get_logger, tenacity_decorator

logger = get_logger(__name__)
MAX_CONTEXT_SIZE = 32768


class ThinkMode(Enum):
    think = "/think"
    not_think = "/no_think"

    @property
    def client_kwargs(self):
        if self == ThinkMode.think:
            return {"temperature": 0.6, "top_p": 0.95}
        else:
            return {}
            # return {"temperature": 0.7, "top_p": 0.8}


@dataclass
class LLM:
    """
    A wrapper class to interact with a language model.
    """

    model: str
    base_url: str | None = None
    api_key: str | None = None
    timeout: int = 360

    def __post_init__(self):
        self.client = OpenAI(
            base_url=self.base_url, api_key=self.api_key, timeout=self.timeout
        )

    @tenacity_decorator
    def __call__(
        self,
        content: str,
        images: str | list[str] | None = None,
        system_message: str | None = None,
        history: list | None = None,
        return_json: bool = False,
        return_message: bool = False,
        think_mode: ThinkMode = ThinkMode.not_think,
        response_format: BaseModel | None = None,
        **client_kwargs,
    ) -> str | dict | list | tuple:
        """
        Call the language model with a prompt and optional images.

        Args:
            content (str): The prompt content.
            images (str or list[str]): An image file path or list of image file paths.
            system_message (str): The system message.
            history (list): The conversation history.
            return_json (bool): Whether to return the response as JSON.
            return_message (bool): Whether to return the message.
            **client_kwargs: Additional keyword arguments to pass to the client.

        Returns:
            Union[str, Dict, List, Tuple]: The response from the model.
        """
        if "qwen3" in self.model.lower():
            client_kwargs.update(think_mode.client_kwargs)
        if history is None:
            history = []
        system, message = self.format_message(
            content, think_mode, images, system_message
        )
        try:
            if response_format is not None:
                completion: ChatCompletion = self.client.chat.completions.parse(
                    model=self.model,
                    messages=system + history + message,
                    response_format=response_format,
                    **client_kwargs,
                )
            else:
                completion: ChatCompletion = self.client.chat.completions.create(
                    model=self.model,
                    messages=system + history + message,
                    **client_kwargs,
                )

        except Exception as e:
            logger.warning("Error in LLM (%s) service: %s", self.model, e)
            raise e
        response = completion.choices[0].message.content
        message.append({"role": "assistant", "content": response})
        return self.__post_process__(response, message, return_json, return_message)

    def __post_process__(
        self,
        response: str,
        message: list,
        return_json: bool = False,
        return_message: bool = False,
    ) -> str | dict | tuple:
        """
        Process the response based on return options.

        Args:
            response (str): The raw response from the model.
            message (List): The message history.
            return_json (bool): Whether to return the response as JSON.
            return_message (bool): Whether to return the message.

        Returns:
            Union[str, Dict, Tuple]: Processed response.
        """
        response = response.strip()
        if return_json:
            response = get_json_from_response(response)
        if return_message:
            response = (response, message)
        return response

    def __repr__(self) -> str:
        repr_str = f"{self.__class__.__name__}(model={self.model}"
        if self.base_url is not None:
            repr_str += f", base_url={self.base_url}"
        return repr_str + ")"

    def test_connection(self) -> bool:
        """
        Test the connection to the LLM.

        Returns:
            bool: True if connection is successful, False otherwise.
        """
        try:
            self.client.models.list()
            return True
        except Exception as e:
            logger.warning(
                "Connection test failed: %s\nLLM: %s: %s, %s",
                e,
                self.model,
                self.base_url,
                self.api_key,
            )
            return False

    def format_message(
        self,
        content: str,
        think_mode: ThinkMode,
        images: str | list[str] | None = None,
        system_message: str | None = None,
    ) -> tuple[list, list]:
        """
        Format messages for OpenAI server call.

        Args:
            content (str): The prompt content.
            images (str or list[str]): An image file path or list of image file paths.
            system_message (str): The system message.

        Returns:
            Tuple[List, List]: Formatted system and user messages.
        """
        if isinstance(images, str):
            images = [images]
        if len(content) > MAX_CONTEXT_SIZE:
            logger.info(f"Input sequence might be too long: {len(content)}")
        if system_message is None:
            if content.startswith("You are"):
                system_message, content = content.split("\n", 1)
            else:
                system_message = "You are a helpful assistant"
        system = [
            {
                "role": "system",
                "content": [{"type": "text", "text": system_message}],
            }
        ]
        message = [{"role": "user", "content": [{"type": "text", "text": content}]}]
        model_idf = self.model.lower()
        if "qwen3" in model_idf:
            system_message += think_mode.value
        if "qwen3" in model_idf or "deepseek" in model_idf:
            system = []
            message[0]["content"][0]["text"] = (
                system_message + message[0]["content"][0]["text"]
            )
        if images is not None:
            for image in images:
                try:
                    with open(image, "rb") as f:
                        message[0]["content"].append(
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": f"data:image/jpeg;base64,{base64.b64encode(f.read()).decode('utf-8')}"
                                },
                            }
                        )
                except Exception as e:
                    logger.error("Failed to load image %s: %s", image, e)
        return system, message

    def gen_image(self, prompt: str, n: int = 1, **kwargs) -> str:
        """
        Generate an image from a prompt.
        """
        return (
            self.client.images.generate(model=self.model, prompt=prompt, n=n, **kwargs)
            .data[0]
            .b64_json
        )

    def get_embedding(
        self,
        text: str,
        encoding_format: str = "float",
        to_tensor: bool = True,
        **kwargs,
    ) -> torch.Tensor | list[float]:
        """
        Get the embedding of a text.
        """
        result = self.client.embeddings.create(
            model=self.model, input=text, encoding_format=encoding_format, **kwargs
        )
        embeddings = [embedding.embedding for embedding in result.data]
        if to_tensor:
            embeddings = torch.tensor(embeddings)
        return embeddings

    def to_async(self) -> "AsyncLLM":
        """
        Convert the LLM to an asynchronous LLM.
        """
        return AsyncLLM(
            model=self.model,
            base_url=self.base_url,
            api_key=self.api_key,
            timeout=self.timeout,
        )


@dataclass
class AsyncLLM(LLM):
    use_batch: bool = False
    """
    Asynchronous wrapper class for language model interaction.
    """

    def __post_init__(self):
        """
        Initialize the AsyncLLM.

        Args:
            model (str): The model name.
            base_url (str): The base URL for the API.
            api_key (str): API key for authentication. Defaults to environment variable.
        """
        self.client = AsyncOpenAI(
            base_url=self.base_url,
            api_key=self.api_key,
            timeout=self.timeout,
        )
        if self.use_batch:
            self.batch = Auto(
                base_url=self.base_url,
                api_key=self.api_key,
                timeout=self.timeout,
                loglevel=0,
            )

    @tenacity_decorator
    async def __call__(
        self,
        content: str,
        images: str | list[str] | None = None,
        system_message: str | None = None,
        history: list | None = None,
        return_json: bool = False,
        return_message: bool = False,
        think_mode: ThinkMode = ThinkMode.not_think,
        response_format: BaseModel | None = None,
        **client_kwargs,
    ) -> str | dict | tuple:
        """
        Asynchronously call the language model with a prompt and optional images.

        Args:
            content (str): The prompt content.
            images (str or list[str]): An image file path or list of image file paths.
            system_message (str): The system message.
            history (list): The conversation history.
            return_json (bool): Whether to return the response as JSON.
            return_message (bool): Whether to return the message.
            **client_kwargs: Additional keyword arguments to pass to the client.

        Returns:
            Union[str, Dict, List, Tuple]: The response from the model.
        """
        if "qwen3" in self.model.lower():
            client_kwargs.update(think_mode.client_kwargs)
        if self.use_batch and threading.current_thread() is threading.main_thread():
            self.batch = Auto(
                base_url=self.base_url,
                api_key=self.api_key,
                timeout=self.timeout,
                loglevel=0,
            )
        elif self.use_batch:
            logger.warning(
                "Warning: AsyncLLM is not running in the main thread, may cause race condition."
            )
        if history is None:
            history = []
        system, message = self.format_message(
            content, think_mode, images, system_message
        )
        try:
            if self.use_batch:
                await self.batch.add(
                    "chat.completions.create",
                    model=self.model,
                    messages=system + history + message,
                    response_format=response_format,
                    **client_kwargs,
                )
                completion = await self.batch.run()
                if "result" not in completion or len(completion["result"]) != 1:
                    raise ValueError(
                        f"The length of completion result should be 1, but got {completion}.\nRace condition may have occurred if multiple values are returned.\nOr, there was an error in the LLM call, use the synchronous version to check."
                    )
                completion = ChatCompletion(**completion["result"][0])
            else:
                if response_format is None:
                    completion = await self.client.chat.completions.create(
                        model=self.model,
                        messages=system + history + message,
                        **client_kwargs,
                    )
                else:
                    completion = await self.client.chat.completions.parse(
                        model=self.model,
                        messages=system + history + message,
                        response_format=response_format,
                        **client_kwargs,
                    )

        except Exception as e:
            logger.error("Error in AsyncLLM call: %s", e)
            raise e
        response = completion.choices[0].message.content
        message.append({"role": "assistant", "content": response})
        return self.__post_process__(response, message, return_json, return_message)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["client"] = None
        state["batch"] = None
        return state

    def __setstate__(self, state: dict):
        self.__dict__.update(state)
        self.client = AsyncOpenAI(
            base_url=self.base_url,
            api_key=self.api_key,
            timeout=self.timeout,
        )
        self.batch = Auto(
            base_url=self.base_url,
            api_key=self.api_key,
            timeout=self.timeout,
            loglevel=0,
        )

    async def test_connection(self) -> bool:
        """
        Test the connection to the LLM asynchronously.

        Returns:
            bool: True if connection is successful, False otherwise.
        """
        try:
            models = await self.client.models.list()
            return any(model.id == self.model for model in models.data)
        except Exception as e:
            logger.warning(
                "Async connection test failed: %s\nLLM: %s: %s, %s",
                e,
                self.model,
                self.base_url,
                self.api_key,
            )
            return False

    async def gen_image(self, prompt: str, n: int = 1, **kwargs) -> str:
        """
        Generate an image from a prompt asynchronously.

        Args:
            prompt (str): The text prompt to generate an image from.
            n (int): Number of images to generate.
            **kwargs: Additional keyword arguments for image generation.

        Returns:
            str: Base64-encoded image data.
        """
        response = await self.client.images.generate(
            model=self.model, prompt=prompt, n=n, response_format="b64_json", **kwargs
        )
        return response.data[0].b64_json

    async def get_embedding(
        self,
        text: str,
        to_tensor: bool = True,
        **kwargs,
    ) -> torch.Tensor | list[float]:
        """
        Get the embedding of a text asynchronously.

        Args:
            text (str): The text to get embeddings for.
            **kwargs: Additional keyword arguments.

        Returns:
            List[float]: The embedding vector.
        """
        response = await self.client.embeddings.create(
            model=self.model,
            input=text,
            encoding_format="float",
            **kwargs,
        )
        embeddings = [embedding.embedding for embedding in response.data]
        if to_tensor:
            embeddings = torch.tensor(embeddings)
        return embeddings

    def to_sync(self) -> LLM:
        """
        Convert the AsyncLLM to a synchronous LLM.
        """
        return LLM(model=self.model, base_url=self.base_url, api_key=self.api_key)


def get_model_abbr(llms: LLM | list[LLM]) -> str:
    """
    Get abbreviated model names from LLM instances.

    Args:
        llms: A single LLM instance or a list of LLM instances.

    Returns:
        str: Abbreviated model names joined with '+'.
    """
    # Convert single LLM to list for consistent handling
    if isinstance(llms, LLM):
        llms = [llms]

    try:
        # Attempt to extract model names before version numbers
        return "+".join(re.search(r"^(.*?)-\d{2}", llm.model).group(1) for llm in llms)
    except Exception:
        # Fallback: return full model names if pattern matching fails
        return "+".join(llm.model for llm in llms)



import re
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum, auto
from types import MappingProxyType
from typing import Any, ClassVar

from lxml import etree
from pptx.dml.fill import FillFormat
from pptx.dml.line import LineFormat
from pptx.enum.dml import MSO_FILL_TYPE
from pptx.enum.shapes import MSO_SHAPE_TYPE
from pptx.oxml.shapes import ShapeElement as PPTXShapeElement
from pptx.oxml.shapes.connector import CT_Connector
from pptx.parts.slide import SlidePart
from pptx.shapes.base import BaseShape
from pptx.shapes.group import GroupShape as PPTXGroupShape
from pptx.shapes.picture import Picture as PPTXPicture
from pptx.shapes.placeholder import PlaceholderPicture, SlidePlaceholder
from pptx.slide import Slide as PPTXSlide
from pptx.slide import _Background
from pptx.text.text import _Paragraph
from pptx.util import Pt

from os.path import join
from pptagent.utils import (
    Config,
    dict_to_object,
    package_join,
    parse_groupshape,
    parsing_image,
    runs_merge,
)

INDENT = "\t"


def shape_normalize(shape: BaseShape):
    """
    This function is used to filter out those malfunctioned attrs.
    """
    if not shape.has_text_frame:
        return
    for para in shape.text_frame.paragraphs:
        for run in para.runs:
            run.hyperlink.address = None


class ClosureType(Enum):
    """
    Enum for closure types.
    """

    CLONE = auto()
    REPLACE = auto()
    DELETE = auto()
    STYLE = auto()
    MERGE = auto()
    POST_PROCESS = auto()

    def __str__(self):
        return self.name.lower()

    @classmethod
    def to_default_dict(cls):
        return {key: [] for key in cls}


@dataclass
class StyleArg:
    """
    A class to represent style arguments for HTML conversion.
    """

    paragraph_id: bool = True
    element_id: bool = True
    font_style: bool = True
    fill_style: bool = True
    area: bool = False
    size: bool = False
    geometry: bool = False
    show_name: bool = False
    show_image: bool = True
    show_empty: bool = False
    show_content: bool = True
    show_semantic_name: bool = False

    @classmethod
    def all_true(cls) -> "StyleArg":
        """
        Create a StyleArg instance with all options enabled.

        Returns:
            StyleArg: A StyleArg instance with all options enabled.
        """
        return cls(
            area=True,
            size=True,
            geometry=True,
            show_semantic_name=True,
        )


@dataclass
class Fill:
    fill_type: MSO_FILL_TYPE
    fill_str: str | None = None
    fill_xml: str | None = None
    image_path: str | None = None

    @classmethod
    def from_shape(cls, fill: FillFormat | None, part: SlidePart, config: Config):
        if fill is None or fill.type is None or fill.type == MSO_FILL_TYPE.BACKGROUND:
            return cls(MSO_FILL_TYPE.BACKGROUND)

        fill_str = "Fill: " + str(fill.value)
        fill_xml = fill._xPr.xml
        fill_type = fill.type
        image_path = None
        if fill_type == MSO_FILL_TYPE.PICTURE:
            image = part.get_image(fill._fill.rId)
            image_path = join(config.IMAGE_DIR, f"{image.sha1}.{image.ext}")
            image_path = parsing_image(image, image_path)
        return cls(fill_type, fill_str, fill_xml, image_path)

    # We pass an element with fill attribute instead of a fill object because `python-pptx` automatically creates a fill object when accessing this attribute, which would cause inconsistency
    def build(
        self, fill_ele: LineFormat | _Background | BaseShape, part: SlidePart
    ) -> None:
        """
        Build the fill in a shape.
        Args:
            shape (BaseShape): The shape to apply fill to.
            fill_xml (Optional[str]): The fill XML to apply.
        """
        if self.fill_type == MSO_FILL_TYPE.BACKGROUND:
            return
        fill = fill_ele.fill
        if self.fill_type == MSO_FILL_TYPE.PICTURE:
            fill.blip()
            _, rId = part.get_or_add_image_part(self.image_path)
            fill.rId = rId
        else:
            new_element = etree.fromstring(self.fill_xml)
            fill._xPr.getparent().replace(fill._xPr, new_element)


@dataclass
class Line:
    fill: Fill
    line_width: float
    line_dash_style: str | None

    @classmethod
    def from_shape(cls, line: LineFormat | None, part: SlidePart, config: Config):
        line_fill = getattr(line, "fill", None)
        if line_fill is None:
            return cls(Fill(MSO_FILL_TYPE.BACKGROUND, "", None), 0, None)
        fill = Fill.from_shape(line_fill, part, config)
        line_width = line.width
        line_dash_style = line.dash_style
        return cls(fill, line_width, line_dash_style)

    def build(self, line: LineFormat, part: SlidePart) -> None:
        """
        Build the line in a shape.
        """
        if self.fill.fill_type == MSO_FILL_TYPE.BACKGROUND:
            return
        self.fill.build(line, part)
        line.width = self.line_width
        line.dash_style = self.line_dash_style


@dataclass
class Background(Fill):
    shape_idx: int = -1

    @classmethod
    def from_slide(cls, slide: PPTXSlide, config: Config) -> "Background":
        """
        Build the background in a slide.

        Args:
            slide (PPTXSlide): The slide to build the background in.
        """
        background = slide.background
        return cls.from_shape(background.fill, slide.part, config)

    def build(self, slide: PPTXSlide) -> None:
        """
        Build the background in a slide.
        """
        super().build(slide.background, slide.part)

    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the background to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the background.
        """
        raise NotImplementedError("Background to HTML conversion is not implemented")

    @property
    def closures(self) -> list:
        """
        Get the closure for the background.
        """
        return []


@dataclass
class Closure:
    """
    A class to represent a closure that can be applied to a shape.
    """

    closure: Callable[[BaseShape], None]
    paragraph_id: int = -1

    def apply(self, shape: BaseShape) -> None:
        """
        Apply the closure to a shape.

        Args:
            shape (BaseShape): The shape to apply the closure to.
        """
        self.closure(shape)

    def __gt__(self, other: "Closure") -> bool:
        """
        Compare closures based on paragraph_id.

        Args:
            other (Closure): Another closure to compare with.

        Returns:
            bool: True if this closure's paragraph_id is greater than the other's.
        """
        if self.paragraph_id != other.paragraph_id:
            return self.paragraph_id > other.paragraph_id


@dataclass
class Font:
    name: str | None = None
    color: str | None = None
    size: int | None = None
    bold: bool | None = None
    italic: bool | None = None
    underline: bool | None = None
    strikethrough: bool | None = None

    def update(self, other: "Font"):
        """
        Merge a list of fonts into a single font.
        """
        for key, value in other.__dict__.items():
            if getattr(self, key) is None:
                setattr(self, key, value)

    def override(self, other: "Font"):
        """
        Merge a list of fonts into a single font.
        """
        for key, value in other.__dict__.items():
            if value is not None:
                setattr(self, key, value)

    def unify(self, others: list["Font"], clear_others: bool = False):
        """
        Merge a list of fonts into a single font.
        """
        if len(others) == 0:
            return
        for key in list(self.__dict__.keys()):
            values = [d.__dict__[key] for d in others]
            if not all(value == values[0] for value in values):
                continue
            setattr(self, key, values[0])
            if not clear_others:
                continue
            for d in others:
                setattr(d, key, None)

    def to_style(self) -> str:
        """
        Convert a font dictionary to a CSS style string.

        Args:
            font (Dict[str, Any]): The font dictionary.

        Returns:
            str: The CSS style string.
        """
        styles = []
        if self.size:
            styles.append(f"font-size: {self.size}pt")

        if self.color:
            styles.append(f"color: #{self.color}")

        if self.bold:
            styles.append("font-weight: bold")

        if self.italic:
            styles.append("font-style: italic")

        return "; ".join(styles)


@dataclass
class Paragraph:
    idx: int
    real_idx: int
    bullet: Any
    font: Font
    text: str
    edited: bool = False

    @classmethod
    def from_paragraph(cls, paragraph: _Paragraph, idx: int) -> "Paragraph":
        run = runs_merge(paragraph)
        real_idx = idx
        bullet = paragraph.bullet
        if run is None:
            return cls(
                idx=-1,
                real_idx=real_idx,
                bullet=bullet,
                font=Font("", "", None, False, False, False, False),
                text="",
            )
        font = Font(**paragraph.font.get_attrs())
        font.override(Font(**run.font.get_attrs()))
        text = re.sub(r"(_x000B_|\\x0b)", " ", paragraph.text)
        return cls(
            idx=idx,
            real_idx=real_idx,
            bullet=bullet,
            font=font,
            text=text,
        )

    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the paragraph to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the paragraph.

        Raises:
            ValueError: If the paragraph is not valid.
        """
        if self.idx == -1:
            raise ValueError(f"paragraph {self.idx} is not valid")
        tag = "li" if self.bullet else "p"
        id_str = f" id='{self.idx}'" if style_args.paragraph_id else ""
        font_style = self.font.to_style()
        style_str = (
            f" style='{font_style}'" if style_args.font_style and font_style else ""
        )
        if self.bullet:
            style_str += f" bullet-type='{self.bullet}'"
        return f"<{tag}{id_str}{style_str}>{self.text}</{tag}>"

    def __repr__(self) -> str:
        """
        Get a string representation of the paragraph.

        Returns:
            str: A string representation of the paragraph.
        """
        return f"Paragraph-{self.idx}: {self.text}"


@dataclass
class TextFrame:
    paragraphs: list[Paragraph] = field(default_factory=list)
    level: int = 0
    text: str = ""
    is_textframe: bool = False
    extents: Any = None
    font: Font | None = None

    @classmethod
    def from_shape(cls, shape: BaseShape, level: int) -> "TextFrame":
        if not shape.has_text_frame:
            return cls(is_textframe=False)
        paragraphs = [
            Paragraph.from_paragraph(paragraph, idx)
            for idx, paragraph in enumerate(shape.text_frame.paragraphs)
        ]
        para_offset = 0
        for para in paragraphs:
            if para.idx == -1:
                para_offset += 1
            else:
                para.idx = para.idx - para_offset
        if len(paragraphs) == 0:
            return cls(is_textframe=False)
        text = shape.text
        extents = shape.text_frame._extents
        try:
            font = Font(**shape.text_frame.font.get_attrs())
        except Exception as _:
            font = Font()
        font.unify([para.font for para in paragraphs if para.idx != -1])
        return cls(
            paragraphs=paragraphs,
            level=level,
            text=text,
            is_textframe=True,
            extents=extents,
            font=font,
        )

    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the text frame to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the text frame.
        """
        if not self.is_textframe:
            return ""
        repr_list = [
            para.to_html(style_args) for para in self.paragraphs if para.idx != -1
        ]
        return "\n".join([INDENT * self.level + repr for repr in repr_list])

    def __repr__(self) -> str:
        if not self.is_textframe:
            return "TextFrame: null"
        return f"TextFrame: {self.paragraphs}"

    def __len__(self) -> int:
        if not self.is_textframe:
            return 0
        return len(self.text)


@dataclass
class ShapeElement:
    config: Config
    slide_idx: int
    shape_idx: int
    style: dict
    data: list
    text_frame: TextFrame
    level: int
    slide_area: float
    sp: PPTXShapeElement
    fill: Fill
    line: Line
    shape: BaseShape | None
    _closures: dict[ClosureType, list[Closure]]

    @classmethod
    def from_shape(
        cls: type["ShapeElement"],
        slide_idx: int,
        shape_idx: int,
        shape: BaseShape,
        config: Config,
        slide_area: float,
        shape_cast: dict[MSO_SHAPE_TYPE, type["ShapeElement"] | None],
        level: int = 0,
    ) -> "ShapeElement":
        """
        Create a ShapeElement from a BaseShape.

        Args:
            slide_idx (int): The index of the slide.
            shape_idx (int): The index of the shape.
            shape (BaseShape): The shape object.
            config (Config): The configuration object.
            slide_area (float): The area of the slide.
            level (int): The indentation level.
            shape_cast (dict[MSO_SHAPE_TYPE, type[ShapeElement]] | None): Optional mapping of shape types to their corresponding ShapeElement classes.
            Set the value to None for any MSO_SHAPE_TYPE to exclude that shape type from processing.
        Returns:
            ShapeElement: The created ShapeElement.

        Raises:
            ValueError: If nested group shapes are not allowed.
        """
        if shape_idx > 100 and isinstance(shape, PPTXGroupShape):
            raise ValueError("Nested group shapes are not allowed")

        shape_normalize(shape)

        # Create style dictionary
        style = {
            "shape_bounds": {
                "width": shape.width,
                "height": shape.height,
                "left": shape.left,
                "top": shape.top,
            },
            "shape_type": str(shape.shape_type).split("(")[0].lower(),
            "rotation": shape.rotation,
            "name": shape.name,
        }

        # Determine semantic name
        try:
            # For auto shapes (rectangle, oval, triangle, star...)
            autoshape = shape.auto_shape_type
            assert autoshape is not None
            style["semantic_name"] = str(autoshape).split()[0].lower().strip()
        except Exception:
            # For other shapes (freeform, connector, table, chart...)
            style["semantic_name"] = str(shape.shape_type).split("(")[0].lower().strip()

        # Create text frame
        text_frame = TextFrame.from_shape(shape, level + 1)

        # Create appropriate shape element based on shape type
        shape_class = shape_cast.get(shape.shape_type, UnsupportedShape)
        if shape_class is UnsupportedShape:
            shape_class = SHAPECAST.get(shape.shape_type, UnsupportedShape)

        if shape_class == Placeholder:
            shape_class = Placeholder.from_shape

        if shape_class == GroupShape:
            shape_class = GroupShape.with_shape_cast(shape_cast)

        return shape_class(
            config=config,
            slide_idx=slide_idx,
            shape_idx=shape_idx,
            style=style,
            data=[],
            text_frame=text_frame,
            level=level,
            slide_area=slide_area,
            sp=deepcopy(shape._element),
            fill=Fill.from_shape(getattr(shape, "fill", None), shape.part, config),
            line=Line.from_shape(getattr(shape, "line", None), shape.part, config),
            shape=shape,
            _closures=ClosureType.to_default_dict(),
        )

    def build(self, slide: PPTXSlide) -> BaseShape:
        """
        Build the shape element in a slide.

        Args:
            slide (PPTXSlide): The slide to build the shape in.

        Returns:
            BaseShape: The built shape.
        """
        sp = deepcopy(self.sp)
        if isinstance(self.sp, CT_Connector):
            sp.nvCxnSpPr.cNvPr.id = slide.shapes._next_shape_id
        else:
            sp.nvSpPr.cNvPr.id = slide.shapes._next_shape_id
        shape = slide.shapes._shape_factory(
            slide.shapes._spTree.insert_element_before(sp, "p:extLst")
        )
        if getattr(shape, "fill", None) is not None:
            self.fill.build(shape, shape.part)
        if getattr(shape, "line", None) is not None:
            self.line.build(shape.line, shape.part)
        return shape

    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the shape element to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the shape element.

        Raises:
            NotImplementedError: If not implemented in a subclass.
        """
        raise NotImplementedError(
            f"to_html not implemented for {self.__class__.__name__}"
        )

    @property
    def text(self) -> str:
        """
        Get the text of the shape element.
        """
        if self.text_frame.is_textframe:
            return self.text_frame.text
        return ""

    def __getstate__(self) -> object:
        state = self.__dict__.copy()
        state["shape"] = None
        return state

    def __repr__(self) -> str:
        """
        Get a string representation of the shape element.

        Returns:
            str: A string representation of the shape element.
        """
        return f"{self.__class__.__name__}: shape {self.shape_idx} of slide {self.slide_idx}"

    @property
    def closures(self) -> list[Closure]:
        """
        Get the closures associated with the shape element.

        Returns:
            List[Closure]: A list of closures.
        """
        closures = []
        closures.extend(sorted(self._closures[ClosureType.CLONE]))
        closures.extend(
            self._closures[ClosureType.REPLACE] + self._closures[ClosureType.STYLE]
        )
        closures.extend(
            sorted(
                self._closures[ClosureType.DELETE]
                + self._closures[ClosureType.POST_PROCESS],
                reverse=True,
            )
        )
        closures.extend(self._closures[ClosureType.MERGE])
        return closures

    @property
    def indent(self) -> str:
        """
        Get the indentation string for the shape element.

        Returns:
            str: The indentation string.
        """
        return "\t" * self.level

    @property
    def left(self) -> float:
        """
        Get the left position of the shape element.

        Returns:
            float: The left position in points.
        """
        return self.style["shape_bounds"]["left"].pt

    @left.setter
    def left(self, value: float) -> None:
        """
        Set the left position of the shape element.

        Args:
            value (float): The left position in points.
        """
        self.style["shape_bounds"]["left"] = value

    @property
    def top(self) -> float:
        """
        Get the top position of the shape element.

        Returns:
            float: The top position in points.
        """
        return self.style["shape_bounds"]["top"].pt

    @top.setter
    def top(self, value: float) -> None:
        """
        Set the top position of the shape element.

        Args:
            value (float): The top position in points.
        """
        self.style["shape_bounds"]["top"] = value

    @property
    def width(self) -> float:
        """
        Get the width of the shape element.

        Returns:
            float: The width in points.
        """
        return self.style["shape_bounds"]["width"].pt

    @width.setter
    def width(self, value: float) -> None:
        """
        Set the width of the shape element.

        Args:
            value (float): The width in points.
        """
        self.style["shape_bounds"]["width"] = value

    @property
    def height(self) -> float:
        """
        Get the height of the shape element.

        Returns:
            float: The height in points.
        """
        return self.style["shape_bounds"]["height"].pt

    @height.setter
    def height(self, value: float) -> None:
        """
        Set the height of the shape element.

        Args:
            value (float): The height in points.
        """
        self.style["shape_bounds"]["height"] = value

    @property
    def area(self) -> float:
        """
        Get the area of the shape element.

        Returns:
            float: The area in square points.
        """
        return self.width * self.height

    @property
    def semantic_name(self) -> str | None:
        """
        Get the semantic name of the shape element.

        Returns:
            Optional[str]: The semantic name, or None if not set.
        """
        return self.style.get("semantic_name", None)

    @semantic_name.setter
    def semantic_name(self, value: str) -> None:
        """
        Set the semantic name of the shape element.

        Args:
            value (str): The semantic name.
        """
        self.style["semantic_name"] = value

    def get_inline_style(self, style_args: StyleArg) -> str:
        """
        Get the inline style for the shape element.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The inline style string.
        """
        id_str = f" id='{self.shape_idx}'" if style_args.element_id else ""
        data_attrs = []
        styles = []

        # Add data attributes
        if style_args.area:
            data_attrs.append(
                f"data-relative-area={self.area * 100 / self.slide_area:.2f}%;"
            )
        if style_args.show_name:
            data_attrs.append(f"data-shapeName='{self.style['name']}'")
        if style_args.show_semantic_name and self.semantic_name is not None:
            data_attrs.append(f"data-semanticName='{self.semantic_name}'")

        # Add style attributes
        if style_args.size:
            styles.append(f"width: {self.width}pt; height: {self.height}pt;")
        if style_args.geometry:
            styles.append(f"left: {self.left}pt; top: {self.top}pt;")
        if style_args.font_style and self.text_frame.is_textframe:
            font_style = self.text_frame.font.to_style()
            if font_style:
                styles.append(font_style)

        # Combine attributes
        if len(styles) != 0:
            id_str += " style='" + " ".join(styles) + "'"
        if len(data_attrs) != 0:
            id_str += " " + " ".join(data_attrs)

        return id_str


@dataclass
class UnsupportedShape(ShapeElement):
    def __post_init__(self) -> None:
        """
        Initialize an UnsupportedShape.

        Raises:
            ValueError: Always, as the shape is unsupported.
        """
        raise ValueError(f"Unsupported shape {self.shape.shape_type}")


class TextBox(ShapeElement):
    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the text box to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the text box.
        """
        content = self.text_frame.to_html(style_args)
        if not style_args.show_content:
            content = ""
        if not content and not style_args.show_empty:
            return ""
        return (
            f"{self.indent}<div{self.get_inline_style(style_args)}>\n"
            + content
            + f"\n{self.indent}</div>\n"
        )


@dataclass
class Picture(ShapeElement):
    """
    A class to represent a picture shape.
    """

    def __post_init__(self):
        """
        Create a Picture from a PPTXPicture.

        Returns:
            Picture: The created Picture.

        Raises:
            ValueError: If the image type is unsupported.
        """
        img_path = join(
            self.config.IMAGE_DIR,
            f"{self.shape.image.sha1}.{self.shape.image.ext}",
        )
        img_path = parsing_image(self.shape.image, img_path)

        # Add image style information
        self.style["img_style"] = {
            "crop_bottom": self.shape.crop_bottom,
            "crop_top": self.shape.crop_top,
            "crop_left": self.shape.crop_left,
            "crop_right": self.shape.crop_right,
        }
        self.data.extend([img_path, self.shape.name, None])  # [img_path, name, caption]

    def build(self, slide: PPTXSlide) -> PPTXPicture:
        """
        Build the picture in a slide.

        Args:
            slide (PPTXSlide): The slide to build the picture in.

        Returns:
            PPTXPicture: The built picture.
        """
        # Add picture to slide
        if self.is_table:
            self.width = Pt(self.width - 1)
            return slide.shapes.add_table(
                self.row, self.col, **self.style["shape_bounds"]
            )

        shape = slide.shapes.add_picture(
            self.img_path,
            **self.style["shape_bounds"],
        )

        # Set properties
        shape.name = self.style["name"]
        dict_to_object(self.style["img_style"], shape.image)

        # Apply shape bounds and rotation
        dict_to_object(self.style["shape_bounds"], shape)
        if hasattr(shape, "rotation"):
            shape.rotation = self.style["rotation"]

        return shape

    @property
    def is_table(self) -> bool:
        return self.style.get("is_table", False)

    @is_table.setter
    def is_table(self, value: bool) -> None:
        self.style["is_table"] = value

    @property
    def grid(self) -> tuple[int, int]:
        assert self.is_table, "The shape is not a table."
        return self.row, self.col

    @grid.setter
    def grid(self, value: tuple[int, int]) -> None:
        assert self.is_table, "The shape is not a table."
        self.row, self.col = value

    @property
    def img_path(self) -> str:
        """
        Get the image path.

        Returns:
            str: The image path.
        """
        return self.data[0]

    @img_path.setter
    def img_path(self, img_path: str) -> None:
        """
        Set the image path.

        Args:
            img_path (str): The image path.
        """
        self.data[0] = img_path

    @property
    def caption(self) -> str | None:
        """
        Get the caption.

        Returns:
            Optional[str]: The caption, or None if not set.
        """
        return self.data[2]

    @caption.setter
    def caption(self, caption: str) -> None:
        """
        Set the caption.

        Args:
            caption (str): The caption.
        """
        self.data[2] = caption

    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the picture to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the picture.

        Raises:
            ValueError: If the caption is not found.
        """
        if not style_args.show_image:
            return ""
        if self.caption is None:
            raise ValueError(
                f"Caption not found for picture {self.shape_idx} of slide {self.slide_idx}"
            )
        return (
            self.indent
            + f"<img {self.get_inline_style(style_args)} alt='{self.caption}'/>"
        )


@dataclass
class GroupShape(ShapeElement):
    """
    A class to represent a group shape.
    """

    shape_cast: ClassVar[dict[MSO_SHAPE_TYPE, type[ShapeElement]]] = {}

    @classmethod
    def with_shape_cast(cls, shape_cast: dict[MSO_SHAPE_TYPE, type[ShapeElement]]):
        """
        Dynamically create a subclass of GroupShape with an isolated shape_cast.
        """
        new_cls = type(f"{cls.__name__}_Isolated_{id(shape_cast)}", (cls,), {})
        new_cls.shape_cast = MappingProxyType(shape_cast)
        return new_cls

    def __post_init__(self) -> None:
        """
        Initialize a GroupShape.
        """
        # Create shape elements for each shape in the group
        self.data = [
            ShapeElement.from_shape(
                self.slide_idx,
                (self.shape_idx + 1) * 100 + i,
                sub_shape,
                self.config,
                self.slide_area,
                self.shape_cast,
                level=self.level + 1,
            )
            for i, sub_shape in enumerate(self.shape.shapes)
            if self.shape_cast.get(sub_shape.shape_type, -1) is not None
            and sub_shape.visible
        ]

        # Apply shape bounds to each shape in the group
        for idx, shape_bounds in enumerate(parse_groupshape(self.shape)):
            if not self.shape.shapes[idx].visible:
                continue
            if self.shape_cast.get(self.shape.shapes[idx].shape_type, -1) is None:
                continue
            self.data[idx].style["shape_bounds"] = shape_bounds

    def build(self, slide: PPTXSlide) -> PPTXSlide:
        """
        Build the group shape in a slide.

        Args:
            slide (PPTXSlide): The slide to build the group shape in.

        Returns:
            PPTXSlide: The slide with the built group shape.
        """
        for shape in self.data:
            shape.build(slide)
        return slide

    def shape_filter(
        self, shape_type: type["ShapeElement"], return_father: bool = False
    ):
        """
        Iterate over all shapes in the group.

        Yields:
            ShapeElement: Each shape in the group.
        """
        for shape in self.data:
            if isinstance(shape, shape_type):
                if return_father:
                    yield (self, shape)
                else:
                    yield shape

    @property
    def shapes(self):
        return self.data

    def __eq__(self, __value: object) -> bool:
        """
        Check if two group shapes are equal.

        Args:
            __value (object): The object to compare with.

        Returns:
            bool: True if the group shapes are equal, False otherwise.
        """
        if not isinstance(__value, GroupShape) or len(self.data) != len(__value.data):
            return False
        for shape1, shape2 in zip(self.data, __value.data):
            if isinstance(shape1, type(shape2)):
                return False
        return True

    def __repr__(self) -> str:
        """
        Get a string representation of the group shape.

        Returns:
            str: A string representation of the group shape.
        """
        return f"{self.__class__.__name__}: {self.data}"

    def __iter__(self):
        return iter(self.data)

    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the group shape to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the group shape.
        """
        content = "\n".join([shape.to_html(style_args) for shape in self.data])
        if not style_args.show_content:
            content = ""
        return (
            self.indent
            + f"<div {self.get_inline_style(style_args)} data-group-label='{self.group_label}'>\n"
            + content
            + "\n"
            + self.indent
            + "</div>\n"
        )

    @property
    def group_label(self) -> str:
        """
        Get the group label.

        Returns:
            str: The group label.
        """
        return getattr(self, "_group_label", f"group_{self.shape_idx}")

    @group_label.setter
    def group_label(self, value: str) -> None:
        """
        Set the group label.

        Args:
            value (str): The group label.
        """
        self._group_label = value


class FreeShape(ShapeElement):
    def to_html(self, style_args: StyleArg) -> str:
        """
        Convert the free shape to HTML.

        Args:
            style_args (StyleArg): The style arguments for HTML conversion.

        Returns:
            str: The HTML representation of the free shape.
        """
        content = self.text_frame.to_html(style_args)
        if not content and not style_args.show_empty:
            return ""
        return (
            f"{self.indent}<div {self.get_inline_style(style_args)}>"
            + f"\n{content}"
            + f"\n{self.indent}</div>"
        )


@dataclass
class SemanticPicture(Picture):
    """
    A class to represent a semantic picture (table, chart, etc.).
    """

    def __post_init__(self):
        shape_type = str(self.shape.shape_type).split()[0]
        self.style["img_style"] = {}
        self.data = [
            package_join("resource", "pic_placeholder.png"),
            self.shape.name,
            f"This is a picture of {shape_type}",
        ]
        self.semantic_name = shape_type


class Placeholder:
    """
    A class to represent a placeholder shape.
    """

    @classmethod
    def from_shape(
        cls,
        shape: SlidePlaceholder,
        **kwargs,
    ) -> Picture | TextBox:
        """
        Create a Placeholder from a SlidePlaceholder.

        Returns:
            Union[Picture, TextBox]: The created shape element.

        Raises:
            ValueError: If the placeholder type is unsupported.
            AssertionError: If the placeholder has multiple types.
        """
        # Ensure placeholder has only one type
        assert (
            sum(
                [
                    shape.has_text_frame,
                    shape.has_chart,
                    shape.has_table,
                    isinstance(shape, PlaceholderPicture),
                ]
            )
            == 1
        ), "Placeholder should have only one type"

        # Create appropriate shape based on placeholder type
        if isinstance(shape, PlaceholderPicture):
            return Picture(
                shape=shape,
                **kwargs,
            )
        elif shape.has_text_frame:
            return TextBox(
                shape=shape,
                **kwargs,
            )
        else:
            raise ValueError(f"Unsupported placeholder {shape.placeholder_type}")


# Define shape type mapping
SHAPECAST = {
    MSO_SHAPE_TYPE.AUTO_SHAPE: FreeShape,
    MSO_SHAPE_TYPE.LINE: FreeShape,
    MSO_SHAPE_TYPE.FREEFORM: FreeShape,
    MSO_SHAPE_TYPE.PICTURE: Picture,
    MSO_SHAPE_TYPE.PLACEHOLDER: Placeholder,
    MSO_SHAPE_TYPE.GROUP: GroupShape,
    MSO_SHAPE_TYPE.TEXT_BOX: TextBox,
    MSO_SHAPE_TYPE.MEDIA: SemanticPicture,
    MSO_SHAPE_TYPE.TABLE: SemanticPicture,
    MSO_SHAPE_TYPE.CHART: SemanticPicture,
    MSO_SHAPE_TYPE.LINKED_PICTURE: SemanticPicture,
    MSO_SHAPE_TYPE.EMBEDDED_OLE_OBJECT: SemanticPicture,
    MSO_SHAPE_TYPE.LINKED_OLE_OBJECT: SemanticPicture,
    MSO_SHAPE_TYPE.DIAGRAM: SemanticPicture,
    MSO_SHAPE_TYPE.CANVAS: SemanticPicture,
    MSO_SHAPE_TYPE.INK: SemanticPicture,
    MSO_SHAPE_TYPE.IGX_GRAPHIC: SemanticPicture,
    MSO_SHAPE_TYPE.WEB_VIDEO: SemanticPicture,
}
